c     ==============================================================
c            MONTE CARLO PROGRAM FOR DUST CONTINUUM TRANSFER
c
c     This program is based on the method of Kenneth Wood. 
c
c     NEWSTUFF 10-11-05:  Possibility for finite-size star.
c     NEWSTUFF 02-02-07: Implemented a much faster method!!
c     NEWSTUFF 02-02-07: Removed imethod=0 and imethod=1...
c     NEWSTUFF 07.05.08: Cleaned up error messages. See tag
c                        [CLEANING ERROR MESSAGES 05.08]
c     ==============================================================

#include "main.h"

#ifndef MIRROR_THETA
      error
#endif
#define COSTEPS 1d-6
#define MINIEPS 1d-12
c#define WRITE_PATH
#define WRITE_STATS
#define TMPPRECREL 1d-4
#define FRMT 1
c#define FRMT 1,*

#ifdef INCLUDE_DUST

c     --------------------------------------------------------------
c                     MAIN ROUTINE FOR MONTE CARLO
c
c     This is the main routine for the Monte Carlo code ala Wood.
c     This routine follows the motion of nphot photon packages as
c     they emerge from the central star and move through the disk.
c     These photons scatter and absorb/reemit multiple times until
c     they escape the system. Each absorb/reemit event also means
c     updating the temperature of the cell, and a change in frequency
c     of the photon package. The temperature of the cell is updated 
c     in such a way that all the photon packages before+including 
c     this one in total have a correct distribution over frequency.
c     This is the trick first described by Kenneth Wood. This 
c     subroutine return the spectrum of the outcoming photons as
c     a function of frequency and inclination, and the temperature
c     of the dust as a function of position. The opacities are all
c     taken from the common_dust.h blocks. 
c
c     --------------------------------------------------------------
      subroutine do_monte_carlo(nphot,spectrum,tdust,scatsrc,miquant,
     %                          meanint,imethod,ifast,enthres,
     %                          iseed,cntdump,irestart,
     %                          ntemp,temp0,temp1)
      implicit none
c
      integer nphot,iseed,imethod,irestart,cntdump,ifast,ntemp
      doubleprecision enthres,temp0,temp1
      doubleprecision spectrum(FRSIZE_FREQ,FRSIZE_Y_SMALL)
      doubleprecision tdust(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                      FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision scatsrc(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision miquant(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
#     ifdef SAVE_MEANINT
      doubleprecision meanint(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
#     else
      doubleprecision meanint
#     endif
c
      doubleprecision starlum(FRSIZE_FREQ)
      doubleprecision extlum(FRSIZE_FREQ)
      doubleprecision energy,theta,mirfact,cumen,tempav
      doubleprecision tmin,tmax,tprec,zzbrent,aa,ran2
      doubleprecision qqum,emcum,rstar
      doubleprecision compute_dusttemp_energy_bd
      doubleprecision compute_dusttemp_coupled_bd
      logical error,ievenodd
      integer integerspec(FRSIZE_FREQ,FRSIZE_Y_SMALL)
      integer inu,imirt,ir,it,nr,nt,iphot,count,countwrite,isize,ispec
      integer iseeddum,cnt,isd,ipstart
c
      doubleprecision pi,pihalf,twopi,fourpi
      parameter(pi=3.14159265358979323846264338328d0)
      parameter(pihalf=1.57079632679489661923132169164d0)
      parameter(twopi=6.28318530717958647692528676656d0)
      parameter(fourpi=12.5663706143591729538505735331d0)
c
      common/abshandle/fnu_old,fnu_diff,cellalpha,ptclener,cellvol
      doubleprecision fnu_old(FRSIZE_FREQ),fnu_diff(FRSIZE_FREQ)
      doubleprecision cellalpha(FRSIZE_FREQ)
      doubleprecision cellvol,ptclener
c
      doubleprecision parsec
      parameter(parsec=3.08572d18)
c
#include "common_grid.h"
#include "common_boundary.h"
#include "common_dust.h"
#include "common_montecarlo.h"
#include "common_montecarlo_arrays.h"
c
c     Do check
c
      if((imethod.eq.0).or.(imethod.eq.1)) then
          write(*,*) 'ERROR: imethod=0 and imethod=1 are obsolete.'
          stop
      endif
      if(istarsurf.eq.1) then
          rstar = radbnd_rstar   ! Starlight is emitted from a sphere (new)
      else
          rstar = 0.d0           ! Starlight is emitted perfectly radially (old)
      endif
c
c     NEW QUANTUM: 28.04.06
c
      if(imethod.ge.10) then
#         ifndef INCLUDE_QUANTSOURCENEW
          write(*,*) 'ERROR: Using imethod.ge.10 you must compile'
          write(*,*) '       with INCLUDE_QUANTSOURCENEW.'
          stop 91280
#         endif
      endif
c
c     Set defaults
c
      countwrite = 1000
      ievenodd   = .true.
      nrskips    = 0
c
c     Define imirt and nr and nt
c
      nr    = irsi_frsizex
#     ifdef MIRROR_THETA
        nt    = (irsi_frsizey+1)/2
        imirt = 1
        mirfact = 0.5
#     else
        nt = irsi_frsizey
        imirt = 0
        mirfact = 1.d0
#     endif
c
c     If write statistics, then open this file
c
#     ifdef WRITE_STATS
      open(unit=4,file='statistics.dat',status='unknown')
      write(4,*) freq_nr
      write(4,*)
#     endif
c
c     Prepare the Monte Carlo simulation 
c
      call prep_monte_carlo(nr,nt,imirt,rrigrid,ttigrid,vol,
     %                      alpha_a,alpha_s,kappa_a,kappa_s,
     %                      tdust,scatsrc,miquant,meanint,
     %                      cumulener,cumulener_bk,
     %                      freqdistr,integerspec,
     %                      ilastphot,iphotcount,
     %                      iphotcurr)
c
c     Prepare the database for the emissivity 
c
      call make_temperature_grid(ntemp,temp0,temp1)
      call make_emiss_dbase()
c
c     Determine the luminosity of the star in all these wavelength
c     bins
c
      do inu=1,freq_nr
          starlum(inu) = pi * radbnd_starspec(inu) *
     %                      fourpi * radbnd_rstar**2
      enddo
      radbnd_starlumtot = 0.d0
      do inu=1,freq_nr
          radbnd_starlumtot = radbnd_starlumtot + 
     %          starlum(inu) * dust_freq_wgt(inu)
      enddo
c
c     Determine the luminosity of the external radiation field.
c     This luminosity is defined as the integral of the incoming
c     part of the external flux at the outer boundary. Of course,
c     if you take the outer boundary further out (for the same
c     problem), then the external luminosity increases, but most
c     of these photons will then simply escape to infinity again
c     without even interacting with the matter. Therefore this
c     scaling is not a major problem. One should keep in mind, 
c     though, that it may strain the photon-statistics when you
c     have too larger outer boundary (photons then rarely reach
c     the inner zones). 
c
c     NEWSTUFF 30-04-04
c
      do inu=1,freq_nr
          extlum(inu) = pi * radbnd_extspec(inu) *
     %                     fourpi * rrigrid(nr+1)**2
      enddo
      radbnd_extlumtot = 0.d0
      do inu=1,freq_nr
          radbnd_extlumtot = radbnd_extlumtot + 
     %          extlum(inu) * dust_freq_wgt(inu)
      enddo
c
c     Compute the luminosity of the internally produced energy qplus
c
c     NEWSTUFF 05-06-04
c
      qpltot = 0.d0
      do ir=1,nr  
          do it=1,nt 
              qpltot = qpltot + qplus(it,ir)*vol(it,ir)
          enddo
      enddo
      if(qpltot.gt.0.d0) then 
          qqum = 0.d0
          do ir=1,nr
              do it=1,nt
                  qplfrac((ir-1)*nt+it) = qqum
                  qqum = qqum + qplus(it,ir)*vol(it,ir)/qpltot
              enddo
          enddo
          qplfrac(nr*nt+1) = qqum
          if(abs(qqum-1.d0).gt.1d-6) then
              write(*,*) 'INTERNAL ERROR: qplus somehow doesnt add up'
              stop
          endif
      endif
c
c     Compute the luminosity of the locally produced spectral input emisloc.
c
c     NEWSTUFF 12-12-04
c
      emisloctot = 0.d0
#     ifdef INCLUDE_EMISLOC
      do ir=1,nr
          do it=1,nt
              do inu=1,freq_nr
                  emisloctot = emisloctot + emisloc(inu,it,ir) *
     %                  fourpi * vol(it,ir) * dust_freq_wgt(inu)
              enddo
          enddo
      enddo
      if(emisloctot.gt.0.d0) then 
          emcum = 0.d0
          do inu=1,freq_nr
              do ir=1,nr
                  do it=1,nt
                      emislocfrac((inu-1)*nt*nr+(ir-1)*nt+it) = emcum
                      emcum = emcum + emisloc(inu,it,ir) * fourpi * 
     %                        vol(it,ir) * dust_freq_wgt(inu) /
     %                        emisloctot
                  enddo
              enddo
          enddo
          emislocfrac(freq_nr*nr*nt+1) = emcum
          if(abs(emcum-1.d0).gt.1d-6) then
              write(*,*) 'INTERNAL ERROR: emisloc somehow doesnt add up'
              stop
          endif
      endif
#     endif
c
c     Compute the luminosity of the quantum heated grains
c
c     NEWSTUFF 12-12-04
c
c     NOTE: Only in the original quantum method (NEW QUANTUM: 28.04.06)
c
      emisloctot   = 0.d0
      emisquanttot = 0.d0
#     ifdef INCLUDE_EMISQUANT
      if((iquantum.lt.0).and.(iquantum.gt.-10)) then
          do ir=1,nr
              do it=1,nt
                  do inu=1,freq_nr
                      emisquanttot = emisquanttot + 
     %                     emisquant(inu,it,ir) *
     %                     fourpi * vol(it,ir) * dust_freq_wgt(inu)
                  enddo
              enddo
          enddo
          if(emisquanttot.gt.0.d0) then 
              emcum = 0.d0
              do inu=1,freq_nr
                  do ir=1,nr
                      do it=1,nt
                          emisquantfrac((inu-1)*nt*nr+(ir-1)*nt+it) = 
     %                              emcum
                          emcum = emcum + emisquant(inu,it,ir) * 
     %                        fourpi *
     %                        vol(it,ir) * dust_freq_wgt(inu) /
     %                        emisquanttot
                      enddo
                  enddo
              enddo
              emisquantfrac(freq_nr*nr*nt+1) = emcum
              if(abs(emcum-1.d0).gt.1d-6) then
                  write(*,*) 'INTERNAL ERROR: emisquant somehow ',
     %                 'doesnt add up'
                  stop
              endif
          endif
      endif
#     endif
c
c     Message
c
      if((iquantum.lt.0).and.(iquantum.gt.-10)) then
          write(*,*) 'L_emisquant/L_star = ',
     %                  emisquanttot/radbnd_starlumtot
      endif
c
c     The energy per photon package is Ltot divided by nr of photons.
c     If there is mirroring in the equator, then reduce the
c     luminosity by half, because only half of the
c     luminosity is going into the upper half-sphere
c
c     NEWSTUFF 30-04-04
c     NEWSTUFF 12-12-04
c
      energy = mirfact * ( qpltot + emisloctot + emisquanttot +
     %         radbnd_starlumtot + radbnd_extlumtot ) / nphot
c
c     Open the path file
c
#     ifdef WRITE_PATH
      open(unit=5,file='path.dat',status='unknown')
#     endif
c
c     Count set to 1
c
      count = 1
      cnt = 1
      write(*,*) ' '
      write(*,*) 'Starting Monte Carlo simulation'
      call flush(6)
c
c     Default
c
      ipstart = 1
c
c     Now, if the irestart.ge.1, then read the latest safety dump
c
      if(irestart.ge.1) then
c
c         Read the last safety backup
c
          write(*,*) '****************************************'
          write(*,*) '    !   RESTARTING THE SIMULATION   !   '
          call flush(6)
          call read_safety_backup(ievenodd,nphot,nr,nt,iphot,
     %                imethod,iseeddum,scatsrc,miquant,
     %                cumulener,cumulener_bk,freqdistr,
     %                tdust,integerspec,count)
          write(*,821) iphot
 821      format('     !   AT PHOTON NR ',I9,'      !   ')
c
c         Now reinitialize the random number generator
c
          if(irestart.eq.1) then 
c
c             Rerandomize using the iseed from the save file
c
              write(*,*) '    ( ran2: randomizing using iseed )'
              write(*,*) '    (       from save file          )'
              if(iseeddum.gt.0) then
                  isd=-iseeddum
              else
                  isd=iseeddum
              endif
              aa=ran2(isd)
              iseed=isd
          elseif(irestart.eq.2) then
c
c             Rerandomize using iseed from montecarlo.inp
c              
              write(*,*) '    ( ran2: randomizing using iseed )'
              write(*,*) '    (       from input file         )'
              if(iseed.gt.0) then
                  isd=-iseed
              else
                  isd=iseed
              endif
              aa=ran2(isd)
              iseed=isd
          elseif(irestart.eq.3) then
c
c             Use exactly the same random set as during last save
c             (is already done, so do not need to do anything here)
c
              write(*,*) '    ( ran2: restoring exact random  )'
              write(*,*) '    (       set as during save      )'
              iseed=iseeddum
          else
              write(*,*) 'ERROR: Restart option must be 1,2 or 3'
              stop 13
          endif
          write(*,*) '****************************************'
          ipstart=iphot
          cnt = 0
      endif
c
c     Launch all the photons
c
      do iphot=ipstart,nphot
c
c         For the photon count routines:
c
          iphotcurr = iphot
c
c         Message
c
          if(count.ge.countwrite) then
              write(*,*) 'Photon nr ',iphot
              call flush(6)
              count=0
          endif
c
c         Safety dump?
c
          if((cnt.ge.cntdump).and.(iphot.lt.nphot)) then
              call make_safety_backup(ievenodd,nphot,nr,nt,iphot,
     %                imethod,iseed,scatsrc,miquant,
     %                cumulener,cumulener_bk,
     %                freqdistr,tdust,integerspec,count,ifast)
              cnt=0
          endif
c
c         Call the walk routine
c
          call walk_full_path(nr,nt,imirt,rrigrid,ttigrid,vol,
     %                alpha_a,alpha_s,kappa_a,rstar,
     %                tdust,scatsrc,miquant,meanint,
     %                cumulener,cumulener_bk,
     %                freqdistr,energy,inu,theta,
     %                imethod,ifast,enthres,iseed)
c
c         If theta > 0, then no error occured
c
          if(theta.ge.0.d0) then
c
c             Find the itheta 
c     
              call hunt(ttigrid,nt,theta,it)
              if(it.gt.nt) it=nt
              if(it.lt.1) it=1
c
c             Add to the outgoing spectrum
c
              integerspec(inu,it) = integerspec(inu,it) + 1
c
          endif
c
c         Update count
c
          count = count + 1
          cnt   = cnt + 1
      enddo
c
c     Close path file
c
#     ifdef WRITE_PATH
      close(5)
#     endif
c
c     If statistics writing active, close
c
#     ifdef WRITE_STATS
      close(4)
#     endif
c
c     Write to the radmc_save.info that we're done
c
      open(unit=1,file='radmc_save.info')
      write(1,*) '-'
      close(1)
c
c     Compute the actual spectrum from the integer spectrum
c
      do it=1,nt
          do inu=1,freq_nr
              spectrum(inu,it) = 
     %           integerspec(inu,it) * energy /
     %          ( twopi * parsec**2 * dust_freq_wgt(inu) * 
     %          abs( cos(ttigrid(it+1)) - cos(ttigrid(it)) ) )
          enddo
      enddo
c
c     Compute the actual tdust
c
      if(itempdecoup.eq.1) then
c
c         If temperatures are decoupled...
c
          do ir=1,nr
              do it=1,nt
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          energy = cumulener(isize,ispec,it,ir) /
     %                         ( dust_rho(ispec,it,ir) *
     %                         vol(it,ir) )
                          tdust(isize,ispec,it,ir) =
     %                         compute_dusttemp_energy_bd(
     %                         energy,ispec,isize)
                      enddo
                  enddo
              enddo
          enddo
      else
c
c         If temperatures are coupled...
c
          do ir=1,nr
              do it=1,nt
                  tempav = compute_dusttemp_coupled_bd(
     %                 cumulener(1,1,it,ir),
     %                 dust_rho(1,it,ir),vol(it,ir))
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          tdust(isize,ispec,it,ir) = tempav
                      enddo
                  enddo
              enddo
          enddo
      endif
c
      end






c     --------------------------------------------------------------
c                     PREPARE THE MONTE CARLO 
c
c     This routine prepares the grid and the arrays for the monte
c     carlo scattering module. This routine must be called for 
c     each frequency-bin separately!!
c     --------------------------------------------------------------
      subroutine prep_monte_carlo(nr,nt,imirt,rrigrid,ttigrid,vol,
     %                            alpha_a,alpha_s,kappa_a,kappa_s,
     %                            tdust,scatsrc,miquant,meanint,
     %                            cumulener,cumulener_bk,freqdistr,
     %                            integerspec,ilastphot,iphotcount,
     %                            iphotcurr)
      implicit none
      integer nr,nt,imirt
      doubleprecision rrigrid(FRSIZE_X+1)
      doubleprecision ttigrid(FRSIZE_Y_SMALL+1)
      doubleprecision tdust(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                      FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision scatsrc(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision miquant(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
#     ifdef SAVE_MEANINT
      doubleprecision meanint(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
#     else
      doubleprecision meanint
#     endif
      doubleprecision alpha_a(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision alpha_s(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision kappa_a(FRSIZE_FREQ,DUST_SIZE_MAX,
     %                        DUST_SPECIES_MAX)
      doubleprecision kappa_s(FRSIZE_FREQ,DUST_SIZE_MAX,
     %                        DUST_SPECIES_MAX)
      doubleprecision vol(FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision cumulener(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision cumulener_bk(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision freqdistr(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      integer integerspec(FRSIZE_FREQ,FRSIZE_Y_SMALL)
      integer ilastphot(FRSIZE_Y_SMALL,FRSIZE_X)
      integer iphotcount(FRSIZE_Y_SMALL,FRSIZE_X)
      integer iphotcurr
c
      doubleprecision temp,find_dust_kappa
      integer ir,it,isize,ispec,inu
c
      doubleprecision pi,pihalf,twopi,fourpi
      parameter(pi=3.14159265358979323846264338328d0)
      parameter(pihalf=1.57079632679489661923132169164d0)
      parameter(twopi=6.28318530717958647692528676656d0)
      parameter(fourpi=12.5663706143591729538505735331d0)
c
#include "common_grid.h"
#include "common_boundary.h"
#include "common_dust.h"
c     
c     Checks
c
      if(dust_opacity_tempdep.ne.0) then
          write(*,*) 'PROBLEM: Temperature dependent opacities'
          write(*,*) '  not yet allowd in Monte Carlo module...'
          stop 13
      endif
c
c     Get the opacities (assume global opacities for the moment)
c
      temp    = 100.d0     ! Arbitrary temperature...
      do inu=1,freq_nr
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  kappa_a(inu,isize,ispec) = 
     %                find_dust_kappa(inu,isize,ispec,temp,1,0)
                  kappa_s(inu,isize,ispec) = 
     %                find_dust_kappa(inu,isize,ispec,temp,0,1)
              enddo
          enddo
      enddo
c
c     Get the grid, and put it into local arrays (easier!)
c
      nr    = irsi_frsizex
#     ifdef MIRROR_THETA
        nt    = (irsi_frsizey+1)/2
        imirt = 1
#     else
        nt = irsi_frsizey
        imirt = 0
#     endif
      do ir=1,nr+1
          rrigrid(ir) = rsi_x_i(ir,1)
      enddo
      do it=1,nt+1
          ttigrid(it) = rsi_x_i(it,2)
      enddo
      if(imirt.eq.1) then
          if(abs(ttigrid(nt+1)-pihalf).gt.1d-3) then
              write(*,*) 'Monte Carlo: ERROR in interface grid!!'
              stop 13
          endif
      endif
c
c     Make sure that the equatorial theta is _EXACTLY_ (30 digits!)
c     equal to pi/2.
c
      if(imirt.eq.0) then
          if(abs(ttigrid(nt/2+1)/pihalf-1.d0).gt.1d-6) stop 91477
          ttigrid((nt+1)/2)=pihalf
      else
          if(abs(ttigrid(nt+1)/pihalf-1.d0).gt.1d-6) then
              write(*,*) ttigrid(nt),pihalf
              stop 91478
          endif
          ttigrid(nt+1)=pihalf
      endif
c
c     Make sure that the polar theta is _EXACTLY_ (30 digits!)
c     equal to 0
c     (Added 15-01-03)
c
      if(abs(ttigrid(1)).gt.1d-6) stop 91477
      ttigrid(1)=0.d0
c
c     Compute the volume elements of the grid cells
c
      do ir=1,nr
          do it=1,nt
              vol(it,ir)  = (1.d0/3.d0) * twopi *
     %               ( rrigrid(ir+1)**3 - rrigrid(ir)**3 ) * 
     %               abs( cos(ttigrid(it)) - cos(ttigrid(it+1)) ) 
          enddo
      enddo
c
c     Make the alpha grids
c
      do ir=1,nr
          do it=1,nt
              do inu=1,freq_nr
                  alpha_a(inu,it,ir) = 0.d0
                  alpha_s(inu,it,ir) = 0.d0
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          alpha_a(inu,it,ir) = alpha_a(inu,it,ir) +
     %                         dust_rho(ispec,it,ir) *
     %                         kappa_a(inu,isize,ispec)
                          alpha_s(inu,it,ir) = alpha_s(inu,it,ir) +
     %                         dust_rho(ispec,it,ir) * 
     %                         kappa_s(inu,isize,ispec)
                      enddo
                  enddo
              enddo
          enddo
      enddo
c
c     Reset the scattering source term
c
      do ir=1,nr
          do it=1,nt
              do inu=1,freq_nr
                  scatsrc(inu,it,ir) = 0.d0
                  miquant(inu,it,ir) = 0.d0
#                 ifdef SAVE_MEANINT
                  meanint(inu,it,ir) = 0.d0
#                 endif
              enddo
          enddo
      enddo
c
c     Reset the tdust and cumulener
c
      do ir=1,nr
          do it=1,nt
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      tdust(isize,ispec,it,ir)        = 0.d0
                      cumulener(isize,ispec,it,ir)    = 0.d0
                      cumulener_bk(isize,ispec,it,ir) = 0.d0
                  enddo
              enddo
          enddo
      enddo
c
c     Reset the freqdistr
c
      do ir=1,nr
          do it=1,nt
              do inu=1,freq_nr
                  freqdistr(inu,it,ir) = 0.d0
              enddo
          enddo
      enddo
c
c     Reset the integer spectrum
c
      do it=1,nt
          do inu=1,freq_nr
              integerspec(inu,it) = 0
          enddo
      enddo
c
c     Reset the photon counters
c
      do ir=1,nr
          do it=1,nt
              ilastphot(it,ir) = 0
              iphotcount(it,ir) = 0
          enddo
      enddo
      iphotcurr = 0
c
      end


c     --------------------------------------------------------------
c         WALK THE ENTIRE RANDOM WALK UNTIL PHOTON LEAVES SYSTEM
c
c     This routine does the scatterings while keeping track of 
c     the energy lost by a package. It only stops after leaving 
c     the grid.
c     --------------------------------------------------------------
      subroutine walk_full_path(nr,nt,imirt,rrigrid,ttigrid,vol,
     %                alpha_a,alpha_s,kappa_a,rstar,
     %                tdust,scatsrc,miquant,meanint,
     %                cumulener,cumulener_bk,
     %                freqdistr,energy,inu,theta,
     %                imethod,ifast,enthres,iseed)
      implicit none
      integer nr,nt,imirt,iseed,inu,imethod,ifast,idestroy
      doubleprecision enthres,rstar
      doubleprecision rrigrid(FRSIZE_X+1)
      doubleprecision ttigrid(FRSIZE_Y_SMALL+1)
      doubleprecision cumulener(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision cumulener_bk(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision freqdistr(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision tdust(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                      FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision scatsrc(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision miquant(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
#     ifdef SAVE_MEANINT
      doubleprecision meanint(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
#     else
      doubleprecision meanint
#     endif
      doubleprecision alpha_a(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision alpha_s(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision kappa_a(FRSIZE_FREQ,DUST_SIZE_MAX,
     %                        DUST_SPECIES_MAX)
      doubleprecision vol(FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision energy
      logical trylin
c
      doubleprecision mu,phi,r,theta,mu0
      doubleprecision ran2,taupath,mu_scat,phi_scat
      doubleprecision mu_out,phi_out,enold
      doubleprecision b,costinf,z00,s00
      doubleprecision z0,s0,dum,x0,sintinf2,epsplus,pi,stot,twopi
      doubleprecision tau,dtau,ds,fr,g,albedo,rn,absorb
      doubleprecision lumtot,lum1,lum2,lum3,lum4
      integer ir,it,isize,ispec,irt,iqactive
      parameter(epsplus=1.d1*TELESC_EPS)
      parameter(pi=PICONST)
      parameter(twopi=6.28318530717958647692528676656d0)
c
#include "common_grid.h"
#include "common_boundary.h"
#include "common_dust.h"
#include "common_montecarlo.h"
c
c     Set trylin
c
      trylin = .false.
      if(ifast.ge.2) then
          trylin = .true.
      endif
c
c     Reset
c
      stot     = 0.d0
      iqactive = 0
c
c     First choose whether to launch the photon from the star or 
c     from the external radiation field or from inside the grid (due
c     to the internally released energy) or from the locally emitted
c     spectral source (emisloc). We make this choice on the 
c     basis of the total luminosity of all four sources.
c
c     NEWSTUFF 30-04-04
c     NEWSTUFF 05-06-04
c     NEWSTUFF 12-12-04
c
      lumtot = radbnd_starlumtot+radbnd_extlumtot+qpltot+
     %         emisloctot+emisquanttot
      lum1   = radbnd_starlumtot
      lum2   = radbnd_starlumtot+radbnd_extlumtot
      lum3   = radbnd_starlumtot+radbnd_extlumtot+qpltot
      lum4   = radbnd_starlumtot+radbnd_extlumtot+qpltot+emisloctot
      rn     = ran2(iseed)
      if(rn.gt.lum1/lumtot) then
          if(rn.gt.lum2/lumtot) then
              if(rn.gt.lum3/lumtot) then
                  if(rn.gt.lum4/lumtot) then
#                     ifdef INCLUDE_EMISQUANT
c
c                     Launch the photon from within the grid due to the 
c                     local emission from PAHs and/or small carbonateous 
c                     grains 
c
c                     NEWSTUFF 12-12-04
c
c                     Check 
c
                      if((iquantum.eq.-10).or.(iquantum.eq.-11)) then
                          stop 23325
                      endif
c
c                     Stupidity check
c
                      if(emisquanttot.eq.0.d0) stop 34992
c
c                     First find the random cell from which to launch this
c                     photon package
c                     
                      rn = ran2(iseed)
c                     
c                     Find the index number 
c                     
                      call hunt(emisquantfrac,freq_nr*nt*nr+1,rn,irt)
c
c                     Find the corresponding index inu, ir and it
c
c##########################MUST CHECK THIS ################
                      inu = (irt-1)/(nr*nt)+1
                      ir  = (irt-(inu-1)*nr*nt-1)/nt+1
                      it  = irt-(ir-1)*nt-(inu-1)*nt*nr
c
c                     Stupidity check
c     
                      if((ir.lt.1).or.(ir.gt.nr).or.
     %                     (it.lt.1).or.(it.gt.nt).or.
     %                     (inu.lt.1).or.(inu.gt.freq_nr)) then
                          write(*,*) 'INTERNAL ERROR: inu, ir ',
     %                               'or it out of range'
                          write(*,*) '    in emisquant picking...'
                          stop
                      endif
c                 
c                     We now know the cell, but now find the position 
c                     within the cell
c
                      call find_position_in_cell(ttigrid,rrigrid,it,ir,
     &                                           r,theta,iseed)
c
c                     Determine random direction
c
                      mu_out  = 2.d0*ran2(iseed)-1.d0
                      phi_out = twopi*ran2(iseed)
                      mu      = mu_out
                      phi     = phi_out
c
c                     Since this is the PAH emission itself, this
c                     must now be treated normally, i.e. not as a
c                     quantum photon
c                     
                      iqactive = 0
c
c                     End of local PAH/quantum launch
c
#                     else
                      write(*,*) 'ERROR: INCLUDE_EMISQUANT is not '
                      write(*,*) '       defined.'
                      write(*,*) '       But somehow local emission '
                      write(*,*) '       is still active...'
                      stop 89028
#                     endif
c
c                     Continue with random walk...
c
                  else
#                     ifdef INCLUDE_EMISLOC
c
c                     Launch the photon from within the grid due to the 
c                     locally emitted radiation from a local stellar
c                     population
c
c                     NEWSTUFF 12-12-04
c
c                     Stupidity check
c
                      if(emisloctot.eq.0.d0) stop 34990
c
c                     First find the random cell from which to launch this
c                     photon package
c                     
                      rn = ran2(iseed)
c                     
c                     Find the index number 
c                     
                      call hunt(emislocfrac,freq_nr*nt*nr+1,rn,irt)
c
c                     Find the corresponding index inu, ir and it
c
c##########################MUST CHECK THIS ################
                      inu = (irt-1)/(nr*nt)+1
                      ir  = (irt-(inu-1)*nr*nt-1)/nt+1
                      it  = irt-(ir-1)*nt-(inu-1)*nt*nr
c
c                     Stupidity check
c     
                      if((ir.lt.1).or.(ir.gt.nr).or.
     %                     (it.lt.1).or.(it.gt.nt).or.
     %                     (inu.lt.1).or.(inu.gt.freq_nr)) then
                          write(*,*) 'INTERNAL ERROR: inu, ir ',
     %                               'or it out of range'
                          write(*,*) '    in emisloc picking...'
                          stop
                      endif
c                 
c                     We now know the cell, but now find the position 
c                     within the cell
c
                      call find_position_in_cell(ttigrid,rrigrid,it,ir,
     &                                           r,theta,iseed)
c
c                     Determine random direction
c
                      mu_out  = 2.d0*ran2(iseed)-1.d0
                      phi_out = twopi*ran2(iseed)
                      mu      = mu_out
                      phi     = phi_out
c
c                     Switch on the quantum flag if quantum heating is on
c                     
                      if((iquantum.lt.0).and.(iquantum.gt.-10)) then
                          iqactive = 1
                      elseif(iquantum.eq.-10) then
                          iqactive = -1
                      elseif(iquantum.eq.-11) then
                          iqactive = 2
                      else
                          iqactive = 0
                      endif
c
c                     End of local launch
c
#                     else
                      write(*,*) 'ERROR: INCLUDE_EMISLOC is not defined'
                      write(*,*) '       But somehow local emission is '
                      write(*,*) '       still active...'
                      stop 89029
#                     endif
c
c                     Continue with random walk...
c
                  endif
              else
c
c                 Launch the photon from within the grid due to the 
c                 internally released energy
c
c                 NEWSTUFF 05-06-04
c
c                 Stupidity check
c
                  if(qpltot.eq.0.d0) stop 34991
c
c                 First find the random cell from which to launch this
c                 photon package
c                 
                  rn = ran2(iseed)
c
c                 Find the index number 
c
                  call hunt(qplfrac,nt*nr+1,rn,irt)
c
c                 Find the corresponding index ir and it
c
                  ir = (irt-1)/nt+1
                  it = irt-(ir-1)*nt
c
c                 Stupidity check
c     
                  if((ir.lt.1).or.(ir.gt.nr).or.
     %                 (it.lt.1).or.(it.gt.nt)) then
                      write(*,*) 'INTERNAL ERROR: ir or it out of range'
                      write(*,*) '                in qplus picking...'
                      stop
                  endif
c
cc########### 
c                 write(*,*) 'RANIDX = ',ir,it,nr,nt,rn,irt
cc###########
c                 
c                 We now know the cell, but now find the position 
c                 within the cell
c
                  call find_position_in_cell(ttigrid,rrigrid,it,ir,
     &                                       r,theta,iseed)
c
c                 Make sure that the local temperature is such that it
c                 could have launched this photon package, and also
c                 at the same time determine the new photon frequency
c                 and direction.
c    
c                 NOTE: In the future it may be important to rethink the
c                       division of this primary energy over the dust
c                       species. I now put inu=-1, which is a signal to
c                       do_asorption_event() to divide the energy by the
c                       density. This may not be correct since some 
c                       species may be much more capable of reemitting
c                       energy than others. For now I simply divide it
c                       per unit dust mass because I have no better idea.
c
                  inu      = -1
                  iqactive = 0
                  call do_absorption_event(vol,dust_rho,kappa_a,alpha_a,
     %                 tdust,cumulener,cumulener_bk,enthres,freqdistr,
     %                 energy,ir,it,inu,1,0,iseed,iqactive,idestroy)
                  mu_out  = 2.d0*ran2(iseed)-1.d0
                  phi_out = twopi*ran2(iseed)
                  mu      = mu_out
                  phi     = phi_out
c
c                 The locally produced heat is usually radiated at 
c                 long wavelengths and thus will not excite PAHs.
c                 So switch off quantum (in old quantum mode).
c                 For new quantum mode, quantum is always active.
c                 NEW QUANTUM: 28.04.06
c
                  if((iquantum.lt.0).and.(iquantum.gt.-10)) then
                      iqactive = 0
                  elseif(iquantum.eq.-10) then
                      iqactive = -1
                  elseif(iquantum.eq.-11) then
                      iqactive = 2
                  else
                      iqactive = 0
                  endif
                  idestroy = 0
c
c                 End of locally released energy
c
c                 Continue with random walk...
c
              endif
          else
c
c             Launch the photon from the outer boundary (i.e. external
c             radiation field)
c
c             We put the photon just inside the outer boundary, with
c             a random Theta position and random directions. 
c     
c             NEWSTUFF 30-04-04
c
              r   = rrigrid(nr+1)*(1.d0-1.d-6)
              rn  = ran2(iseed)
              if(imirt.eq.1) then
                  theta = acos(rn)
              else
                  theta = acos(2*rn-1.d0)
              endif
              rn  = ran2(iseed)
              mu  = -sqrt(rn)   ! Because P(mu)=2*|mu| for influx!
                                ! Re-checked 10-11-05: is OK indeed.
              rn  = ran2(iseed)
              phi = 2*pi*rn
c
c             We determine the frequency of the photon using the spectrum
c             of the external radiation
c
              call randomfreq(radbnd_extspec,inu,iseed)
c
c             Switch on the quantum flag if quantum heating is on
c                     
              if((iquantum.lt.0).and.(iquantum.gt.-10)) then
                  iqactive = 1
              elseif(iquantum.eq.-10) then
                  iqactive = -1
              elseif(iquantum.eq.-11) then
                  iqactive = 2
              else
                  iqactive = 0
              endif
c
c             End of externally produced photons
c
c             Continue with random walk...
c
          endif
      else
c
c         Launch the photon from the central star
c
          if(rstar.eq.0.d0) then
c
c             **** Original method: launch perfectly radially ****
c
c             Switch on the quantum flag if quantum heating is on
c             
              if((iquantum.lt.0).and.(iquantum.gt.-10)) then
                  iqactive = 1
              elseif(iquantum.eq.-10) then
                  iqactive = -1
              elseif(iquantum.eq.-11) then
                  iqactive = 2
              else
                  iqactive = 0
              endif
c
c             First choose a random angle at which the photon package leaves
c             the origin of the coordinate system
c
              mu  = 1.d0
              phi = 0.d0
              r   = 0.d0
              rn  = ran2(iseed)
              if(imirt.eq.1) then
                  theta = acos(rn)
              else
                  theta = acos(2*rn-1.d0)
              endif
c
c             Find a random inu for the photon package, based on the 
c             stellar flux
c
              call randomfreq(radbnd_starspec,inu,iseed)
c             
c             Now do the first leg of the walk: the radial movement away from 
c             the star and into the environment
c             
              r = 1d30
              call huntalt(ttigrid,nt+1,theta,it)
c             
c             Check if photon happens to pass exactly along a theta coordinate
c             
              if(theta.eq.ttigrid(it)) stop 32001
c             
c             Check if the photon is penetrating the grid
c             
              if((it.lt.1).or.(it.gt.nt)) then
                  return
              endif
c             
c             At which tauscat do we have a scattering event?
c             
              taupath = -log(1.d0-ran2(iseed))
c             
c             Follow the path
c             
              tau = 0.d0
              do ir=1,nr
                  ds     = (rrigrid(ir+1)-rrigrid(ir))
                  dtau   = ( alpha_a(inu,it,ir) + 
     %                       alpha_s(inu,it,ir) ) * ds
                  albedo = alpha_s(inu,it,ir)/
     %                 (alpha_a(inu,it,ir)+alpha_s(inu,it,ir))
                  absorb = 1.d0-albedo
c                 
c                 Check if we are already at the scattering or absorption event
c                 
                  if(tau+dtau.gt.taupath) then
c                     
c                     Compute the radial position of this event
c                     
                      fr     = (taupath-tau)/dtau
                      r      = rrigrid(ir) + fr * ds 
c                     
c                     Update stot
c                     
                      stot = stot + fr * ds
c                     
c                     Add this continuous contribution
c                     to the current cell
c                     
                      do ispec=1,dust_nr_species
                          if((dust_quantum(ispec).eq.0).or.
     %                         (iqactive.le.0)) then
                              do isize=1,dust_nr_size(ispec)
                                  cumulener(isize,ispec,it,ir) = 
     %                                 cumulener(isize,ispec,it,ir)+
     %                                 absorb*(taupath-tau)*energy*
     %                                 dust_rho(ispec,it,ir) * 
     %                                 kappa_a(inu,isize,ispec) / 
     %                                 alpha_a(inu,it,ir)
                              enddo
                          endif
                      enddo
c
c                     Add photons to scattering source
c
                      scatsrc(inu,it,ir) = scatsrc(inu,it,ir) +
     %                     albedo*(taupath-tau)*energy/
     %                     (vol(it,ir)*dust_freq_wgt(inu))
c
c                     Add photons to mean intensity 
c                     
#                     ifdef SAVE_MEANINT
                      meanint(inu,it,ir) = meanint(inu,it,ir) +
     %                             (taupath-tau)*energy/
     %                             (vol(it,ir)*dust_freq_wgt(inu)*
     %                              (alpha_a(inu,it,ir)+
     %                               alpha_s(inu,it,ir))*12.566371d0)
#                     endif
c
c                     Add photons to mean intensity of 
c                     primary (quantum-heating) photons
c
c                     NEWSTUFF 12-12-04
c
                      if(iqactive.ne.0) then
                          miquant(inu,it,ir) = miquant(inu,it,ir) +
     %                         (taupath-tau)*energy/
     %                         (vol(it,ir)*dust_freq_wgt(inu)*
     %                         (alpha_a(inu,it,ir)+
     %                         alpha_s(inu,it,ir))*12.566371d0)
                      endif
c
c                     End this first radial trajectory
c                     
                      goto 10
                  endif
c                 
c                 Update stot
c                 
                  stot = stot + ds
c                 
c                 Increase the tau
c                 
                  tau = tau + dtau
c                 
c                 Add this continuous contribution
c                 to the current cell
c                 
                  do ispec=1,dust_nr_species
                      if((dust_quantum(ispec).eq.0).or.
     %                     (iqactive.le.0)) then
                          do isize=1,dust_nr_size(ispec)
                              cumulener(isize,ispec,it,ir) = 
     %                             cumulener(isize,ispec,it,ir) +
     %                             absorb*dtau*energy*
     %                             dust_rho(ispec,it,ir) * 
     %                             kappa_a(inu,isize,ispec) / 
     %                             alpha_a(inu,it,ir)
                          enddo
                      endif
                  enddo
c
c                 Add photons to scattering source
c
                  scatsrc(inu,it,ir) = scatsrc(inu,it,ir) +
     %                 albedo*dtau*energy/
     %                 (vol(it,ir)*dust_freq_wgt(inu))
c
c                 Add photons to mean intensity 
c
#                 ifdef SAVE_MEANINT
                  meanint(inu,it,ir) = meanint(inu,it,ir) +
     %                         dtau*energy/
     %                         (vol(it,ir)*dust_freq_wgt(inu)*
     %                          (alpha_a(inu,it,ir)+
     %                           alpha_s(inu,it,ir))*12.566371d0)
#                 endif
c
c                 Add photons to mean intensity of 
c                 primary (quantum-heating) photons
c
c                 NEWSTUFF 12-12-04
c
                  if(iqactive.ne.0) then
                      miquant(inu,it,ir) = miquant(inu,it,ir) +
     %                     dtau*energy/
     %                     (vol(it,ir)*dust_freq_wgt(inu)*
     %                     (alpha_a(inu,it,ir)+
     %                     alpha_s(inu,it,ir))*12.566371d0)
                  endif
c
c                 Increase the cell count
c                 (NEWSTUFF 27.01.07)
c
                  iphotcount(it,ir) = iphotcount(it,ir) + 1
c                 
              enddo
c             
c             Left the grid from the outside
c             
              r = 1d30
              return
c         
 10           continue
c
c             Decide whether this was a scattering event or an
c             absorption event
c
              if(ran2(iseed).lt.albedo) then
c                 
c                 Scattering event
c             
                  g   = 0.d0    ! Must be improved at a later stage
c##########################
c##########################
c                 call henyeygreenstein(g,mu_scat,phi_scat,iseed)
c                 call rotate(mu,phi,mu_scat,phi_scat,mu_out,phi_out)
c##########################
                  mu_out  = 2.d0*ran2(iseed)-1.d0
                  phi_out = twopi*ran2(iseed)
c##########################
                  mu  = mu_out
                  phi = phi_out
              else
c                 
c                 Absorption event
c                 
                  call do_absorption_event(vol,dust_rho,kappa_a,alpha_a,
     %                 tdust,cumulener,cumulener_bk,enthres,freqdistr,
     %                 energy,ir,it,inu,imethod,ifast,iseed,
     %                 iqactive,idestroy)
                  mu_out  = 2.d0*ran2(iseed)-1.d0
                  phi_out = twopi*ran2(iseed)
                  mu      = mu_out
                  phi     = phi_out
c
c                 Check if photon must be destroyed
c
                  if(idestroy.ne.0) then
                      theta = -1d30 ! Signal that photon is destroyed
                      return
                  endif
c
c                 Switch off the quantum flag 
c             
c                 ONLY for old quantum method
c
                  if(iquantum.gt.-10) then
                      iqactive = 0
                  endif
c
              endif
c
c             End of radially outward launched starlight
c
c             Continue with random walk...
c
          else
c
c             **** Launch photon from real stellar surface ****
c             (NEWSTUFF 10-11-05)
c
c             Switch on the quantum flag if quantum heating is on
c             
              if((iquantum.lt.0).and.(iquantum.gt.-10)) then
                  iqactive = 1
              elseif(iquantum.eq.-10) then
                  iqactive = -1
              elseif(iquantum.eq.-11) then
                  iqactive = 2
              else
                  iqactive = 0
              endif
c
c             First choose a random theta at which the photon from the
c             (spherically symmetric) star enters the inner edge of the
c             coordinate system.
c
              rn  = ran2(iseed)
              if(imirt.eq.1) then
                  theta = acos(rn)
              else
                  theta = acos(2*rn-1.d0)
              endif
c
c             The inner edge of the coordinate system is:
c
              r    = rrigrid(1) + 1d-3 * (rrigrid(2)-rrigrid(1))
c
c             Now at which angle does the photon start here? This 
c             depends on the stellar radius... We first start at
c             the surface of the star, and use the sqrt(rn) to 
c             produce a local mu that is consistent with the 
c             proper photon statistics. Then we transport this outward
c             to the inner edge of the grid.
c
c             NOTE: I have been confused again, here, with this
c             sqrt(rn). But it is indeed correct, as far as I know!!
c             The example that made me understand again why this is so
c             is the example of a transport company with fast trucks,
c             sending out per day 10 trucks, and a firm with slow
c             trucks, also sending out per day 10 trucks. Both firms have
c             the same flux of goods, even though firm2 has slow
c             trucks. Firm2 just requires more trucks to keep the flux
c             going. That is why photons with mu\simeq 0 (slow) would
c             carry the same flux as photons with mu\simeq 1 (fast) if
c             they would be emitted equally often. Since we know that
c             mu\simeq 0 photons carry less flux, they must be emitted
c             less often. Hence P(mu)=mu, i.e. Pcum(mu)=mu^2, or so to
c             say: mu=sqrt(Pcum)=sqrt(rn).
c             
              if(r.gt.1d3*rstar) then
                  mu = 1.d0
              else
                  if(r.le.rstar) stop 28670
                  rn  = ran2(iseed)
                  mu0 = sqrt(rn)
                  mu  = sqrt(1.d0-((rstar/r)**2)*(1.d0-mu0**2))
              endif
c
c             Now the random phi
c
              phi   = twopi*ran2(iseed)
c
c             Find a random inu for the photon package, based on the 
c             stellar flux
c
              call randomfreq(radbnd_starspec,inu,iseed)
c
c             End of stellar surface launched starlight
c             (Finished 10-11-05)
c
c             Continue with random walk...
c
          endif
      endif
c
c     Now continue with the random walk
c
 50   continue
c
c     Get a random tau
c
      taupath = -log(1.d0-ran2(iseed))
c
c     Move photon to next scattering/absorption event
c         
      call walk_cells_twodee(inu,r,theta,mu,phi,
     %                nr,nt,imirt,rrigrid,ttigrid,vol,
     %                alpha_a,alpha_s,kappa_a,taupath,
     %                scatsrc,miquant,meanint,
     %                cumulener,energy,imethod,stot,trylin,
     %                iqactive,idestroy)
c
c     If escaped, then return
c
      if(r.ge.1d29) then
          return
      endif
c
c     If error, then also return
c
      if(r.lt.0.d0) then
          theta=-1d30
          write(*,*) 'Cell interface-crossing algorithm ',
     %               'failed for this photon... Skipping it...'
          call flush(6)
          nrskips = nrskips + 1
          return
      endif
c
c     If photon destroyed (by e.g. quantum heated grains which are
c     dealt with elsewhere), then also return
c
      if(idestroy.ne.0) then
          theta=-1d30
          return
      endif
c
c     Find the ir,it
c
      call huntalt(rrigrid,nr+1,r,ir)
      call huntalt(ttigrid,nt+1,theta,it)
      if((ir.lt.1).or.(ir.gt.nr).or.(it.lt.1).or.(it.gt.nt)) stop 16501
c
c     [BUG FIXED: 26-02-03 (Rowin Meijerink)]
c     Find the new albedo
c
      albedo = alpha_s(inu,it,ir)/
     %     (alpha_a(inu,it,ir)+alpha_s(inu,it,ir))
c
c     Decide whether this was a scattering event or an
c     absorption event
c
      if(ran2(iseed).lt.albedo) then
c
c         Scattering event
c
          g   = 0.d0            ! Must be improved at a later stage
c##########################
c##########################
c          call henyeygreenstein(g,mu_scat,phi_scat,iseed)
c          call rotate(mu,phi,mu_scat,phi_scat,mu_out,phi_out)
c##########################
          mu_out  = 2.d0*ran2(iseed)-1.d0
          phi_out = twopi*ran2(iseed)
c##########################
          mu  = mu_out
          phi = phi_out
      else
c
c         Absorption event
c
          call do_absorption_event(vol,dust_rho,kappa_a,alpha_a,tdust,
     %                    cumulener,cumulener_bk,enthres,freqdistr,
     %                    energy,ir,it,inu,imethod,ifast,iseed,
     %                    iqactive,idestroy)
          mu_out  = 2.d0*ran2(iseed)-1.d0
          phi_out = twopi*ran2(iseed)
          mu      = mu_out
          phi     = phi_out
c
c         If photon destroyed (by e.g. quantum heated grains which are
c         dealt with elsewhere), then also return
c
          if(idestroy.ne.0) then
              theta=-1d30
              return
          endif
c
c         Switch off quantum
c
c         ONLY for old quantum method
c
          if(iquantum.gt.-10) then
              iqactive = 0
          endif
c
      endif
c
c     Next event
c
      goto 50
c
      end




c     --------------------------------------------------------
c                    FOLLOW A PHOTON ALONG A RAY
c
c     This routine follows a photon until the next scattering
c     event. In the mean time it leaves its energy behind by
c     continuous absorption in the cells it crosses. The
c     ray (trajectory) of the photon already must have been
c     set up prior to this call by make_trajectory_c() from
c     the telescope.F module.
c     --------------------------------------------------------
      subroutine walk_cells_twodee(inu,r,theta,mu,phi,
     %                nr,nt,imirt,rrigrid,ttigrid,vol,
     %                alpha_a,alpha_s,kappa_a,taupath,
     %                scatsrc,miquant,meanint,
     %                cumulener,energy,imethod,stot,trylin,
     %                iqactive,idestroy)
      implicit none
      integer nr,nt,imirt,inu,imethod,iqactive,idestroy
      logical trylin
      doubleprecision r,theta,mu,phi,stot,energy
      doubleprecision rold,told,muold,phiold
      doubleprecision rrigrid(FRSIZE_X+1)
      doubleprecision ttigrid(FRSIZE_Y_SMALL+1)
      doubleprecision vol(FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision scatsrc(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision miquant(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
#     ifdef SAVE_MEANINT
      doubleprecision meanint(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
#     else
      doubleprecision meanint
#     endif
      doubleprecision alpha_a(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision alpha_s(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision kappa_a(FRSIZE_FREQ,DUST_SIZE_MAX,
     %                        DUST_SPECIES_MAX)
      doubleprecision cumulener(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision taupath,fr
c
      integer ir1,ircell,it1,itcell,igrid,nteq,isize,ispec
      integer idir,ileftright,is
      doubleprecision r0,r1,t0,t1,mu0,mu1,phi0,phi1
      doubleprecision tau,dtau,sold,s,albedo,absorb
      doubleprecision dexp,dener,costinf,ds,sdum,b,z0,dummy,sinphi
      doubleprecision r_cell,t_cell
c
      doubleprecision pi,pihalf,twopi,fourpi
      parameter(pi=3.14159265358979323846264338328d0)
      parameter(pihalf=1.57079632679489661923132169164d0)
      parameter(twopi=6.28318530717958647692528676656d0)
      parameter(fourpi=12.5663706143591729538505735331d0)
c
#include "common_montecarlo.h"
#include "common_dust.h"
c
c     Check some stuff
c
      if((phi.lt.0.d0).or.(phi.gt.twopi)) stop 77732
      if(r.gt.rrigrid(nr+1)) then
          write(*,*) 'R out of bounds'
          write(*,*) r,rrigrid(1),rrigrid(nr+1)
          write(*,*) ircell,r0,r1
          stop 49201
      endif
      if((theta.gt.ttigrid(nt+1)).or.(theta.lt.ttigrid(1))) then
          write(*,*) 'Theta out of bounds'
          write(*,*) theta,ttigrid(1),ttigrid(nt+1)
          write(*,*) rold,told,muold,phiold
          write(*,*) r,theta,mu,phi
          write(*,*) r0,r1,t0,t1,t1/pihalf-1.d0
          stop 49202
      endif
c
c     For the moment: do not allow non-mirroring.
c
      if(imirt.eq.0) then
          write(*,*) 'PROBLEM: Monte Carlo still only for Mirror Theta'
          stop 13
      endif
c
c     Compute the nt of the equator
c
      if(imirt.eq.1) then
          nteq = nt+1
      else
          nteq = nt/2+1
      endif
c
c     Reset tau
c
      tau   = 0.d0
c
c     Default
c
      idestroy = 0
c
c     If we are at bottom quadrant, and if mirror active, then
c     flip to upper quadrant (although this should in principle
c     not happen!)
c
      if((imirt.eq.1).and.(theta.gt.pihalf)) then
          theta = pi - theta
          phi   = twopi - phi 
          write(*,*) 'FLIPPING NEEDED...'
      endif
c
c     First we are presumably somewhere in a cell. Find the ir,it
c
      call huntalt(rrigrid,nr+1,r,ircell)
      call huntalt(ttigrid,nt+1,theta,itcell)
c
c     Check if indeed we are in a cell
c     [CLEANING ERROR MESSAGES 05.08]
c
      if((ircell.lt.1).or.(ircell.gt.nr).or.
     %   (itcell.lt.1).or.(itcell.gt.nt)) then
          if(itcell.gt.nt) then
              theta = ttigrid(nt+1)-1d-12
              itcell = nt
          else
c             write(*,*) 'ERROR: out of cell somehow...'
             r = -1d30
             return
c              stop 84234
          endif
      endif
c
c     Check for bizarre coincidence that a photon is exactly at a
c     grid boundary
c     [CLEANING ERROR MESSAGES 05.08]
c
      if(r.eq.rrigrid(ircell)) then
c          write(*,*) 'BIZARRE COINCIDENCE: Exactly at r-boundary'
c          write(*,*) 'Fixing...'
          if(ircell.gt.1) then 
              r      = r * (1.d0 - 1d-12)
              ircell = ircell - 1
          else
              r      = r * (1.d0 + 1d-12)
          endif
      endif
      if(theta.eq.ttigrid(itcell)) then
c          write(*,*) 'BIZARRE COINCIDENCE: Exactly at theta-boundary'
c          write(*,*) 'Fixing...'
          if(itcell.gt.1) then 
              theta = theta - 1d-12
              itcell = itcell - 1
          else
              theta = theta + 1d-12
          endif
      endif
c
c     Signal that we are indeed on the grid at this point
c
      igrid = 1
c
c     Increase the cell count
c     (NEWSTUFF 27.01.07)
c
      if(iphotcurr.gt.ilastphot(itcell,ircell)) then
          iphotcount(itcell,ircell) = iphotcount(itcell,ircell) + 1
          ilastphot(itcell,ircell)  = iphotcurr
      endif
c
c     Backup old position
c
      rold   = r
      told   = theta
      muold  = mu
      phiold = phi
c
c     In the fast (linear) mode, we first check if the next event will be
c     within the cell, and only little distance from where we were. In this
c     limit we can use the linear approximation of the trajectory.
c
      if(trylin) then
c
c         What is the closest distance to a neighboring cell-interface?
c
          ds = 1d99
          if(r-rrigrid(ircell).lt.ds) ds = r-rrigrid(ircell)
          if(rrigrid(ircell+1)-r.lt.ds) ds = rrigrid(ircell+1)-r
          if(r*(theta-ttigrid(itcell)).lt.ds) ds = r*
     %                        (theta-ttigrid(itcell))
          if(r*(ttigrid(itcell+1)-theta).lt.ds) ds = r*
     %                        (ttigrid(itcell+1)-theta)
          if(ds.lt.0.d0) stop 22139
c
c         If next event is well within this same cell then we do
c         things linear here
c
          dtau   = ( alpha_a(inu,itcell,ircell) +
     %               alpha_s(inu,itcell,ircell) ) * ds
          dtau   = dtau * 0.1    ! This is the safety margin!
          if(tau+dtau.gt.taupath) then
c
c             Do things linear. Keep mu and phi constant, and update
c             the r and theta according to the linear formulae
c
c             Compute the real ds
c
              ds = taupath / ( alpha_a(inu,itcell,ircell) +
     %                          alpha_s(inu,itcell,ircell) )
c
c             Check
c
              if(alpha_a(inu,itcell,ircell) +
     %           alpha_s(inu,itcell,ircell).eq.0.d0) stop 98230
c
c             Add energy to cell
c
              do ispec=1,dust_nr_species
                  if((dust_quantum(ispec).eq.0).or.
     %                 (iqactive.le.0)) then 
                      do isize=1,dust_nr_size(ispec)
                          cumulener(isize,ispec,itcell,ircell) = 
     %                         cumulener(isize,ispec,itcell,ircell)+
     %                         absorb*taupath*energy*
     %                         dust_rho(ispec,itcell,ircell) * 
     %                         kappa_a(inu,isize,ispec) / 
     %                         alpha_a(inu,itcell,ircell)
                      enddo
                  endif
              enddo
c
c             Add photons to scattering source
c
              scatsrc(inu,itcell,ircell) = 
     %             scatsrc(inu,itcell,ircell) +
     %             albedo*taupath*energy/
     %             (vol(itcell,ircell)*dust_freq_wgt(inu))
c
c             Add photons to mean intensity 
c
#             ifdef SAVE_MEANINT
              meanint(inu,itcell,ircell) = 
     %              meanint(inu,itcell,ircell) +
     %              taupath*energy/
     %              (vol(itcell,ircell)*dust_freq_wgt(inu)*
     %              (alpha_a(inu,itcell,ircell)+
     %               alpha_s(inu,itcell,ircell))*12.566371d0)
#             endif
c
c             Add photons to mean intensity of 
c             primary (quantum-heating) photons
c
c             NEWSTUFF 12-12-04
c
              if(iqactive.ne.0) then
                  miquant(inu,itcell,ircell) = 
     %                 miquant(inu,itcell,ircell) +
     %                 taupath*energy/
     %                 (vol(itcell,ircell)*dust_freq_wgt(inu)*
     %                 (alpha_a(inu,itcell,ircell)+
     %                 alpha_s(inu,itcell,ircell))*12.566371d0)
              endif
c
c             Compute the location of this point
c
c             ... the s
c
              s    = sold + ds
              stot = stot + ds
c
c             ... the r
c
              r   = rold + mu * ds
c
c             ... the theta
c
              theta = told - sin(phiold) * sqrt(1-mu*mu) * ds / rold
c
c             ... the mu
c
              mu  = muold
c
c             ... the phi
c
              phi = phiold
c
c             Check that they are still within the same cell
c             (safety check, as it should not be possible)
c
              if((r.lt.rrigrid(ircell)).or.
     %           (r.gt.rrigrid(ircell+1)).or.
     %           (theta.lt.ttigrid(itcell)).or.
     %           (theta.gt.ttigrid(itcell+1))) then
                  write(*,*) 'MAJOR PROBLEM OCCURRED:'
                  write(*,*) '  Linear photon movement went astray!'
                  write(*,*) '  Aborting...'
                  stop 13
              endif
c
c             Write path element
c
#             ifdef WRITE_PATH
              write(5,*) r,theta,mu,phi,s,tau_s,stot,energy
#             endif
c
              return
          endif
      endif
c
c     Find the next crossing with a cell interface
c     Note that at the start of this walk we must be _inside_ a
c     cell, and we also end _inside_ of a cell. But the other
c     steps are from cell boundary to cell boundary.
c
      r0 = rrigrid(ircell)
      r1 = rrigrid(ircell+1)
      t0 = ttigrid(itcell)
      t1 = ttigrid(itcell+1)
      call find_next_crossing(rold,r0,r1,told,t0,t1,
     %         muold,phiold,sold,r,theta,mu,phi,s,
     %         idir,ileftright,b,costinf,z0)
c
c     Check...
c
      if(r.lt.0.d0) then
c          write(*,*) '(error at position A)'
c          write(0,*) '(error at position A)'
          return
      endif
      if(r.gt.rrigrid(nr+1)) then
          write(*,*) 'R out of bounds'
          write(*,*) r,rrigrid(1),rrigrid(nr+1)
          write(*,*) ircell,r0,r1
          stop 7491
      endif
c
c     Write path element
c
#     ifdef WRITE_PATH
      write(5,*) rold,told,muold,phiold,sold,tau,stot,energy
#     endif
c
c     Do the loop from one cell crossing to the next
c
      do is=1,2*nr+2*nt+2
c
c         If we are at bottom quadrant, and if mirror active, then
c         something must have gone wrong...
c
          if((imirt.eq.1).and.(theta.gt.pihalf)) then
c              write(*,*) 'Internal error (194375). Skipping 1 photon.'
c              write(0,*) 'Internal error (194375). Skipping 1 photon.'
              r       = -1d30
              return
c              stop 194375
          endif
c##########################
          if(r.gt.rrigrid(nr+1)) then
              write(*,*) 'A1 = ',is,r,rrigrid(nr+1)
          endif
c##########################
c
c         Now find the ir1,it1
c
          call huntalt(rrigrid,nr+1,r,ir1)
          call huntalt(ttigrid,nt+1,theta,it1)
c
c         Check...
c
          if(ir1.gt.nr+1) then
              write(*,*) is,r,rrigrid(nr+1)
              stop 39991
          endif
          if(it1.gt.nt+1) then
              write(*,*) is,theta,ttigrid(nt+1)
              stop 39992
          endif
c
c         Safety check...
c
c####################################
          if(igrid.eq.1) then
              if((ircell.lt.1).or.(ircell.gt.nr).or.
     %           (itcell.lt.1).or.(itcell.gt.nt)) then
                  stop 22371
              endif
          else
              if((ircell.ge.1).and.(ircell.le.nr).and.
     %           (itcell.ge.1).and.(itcell.le.nt)) then
                  stop 22372
              endif
          endif
c####################################
c
c         Path length
c
          ds   = s - sold
c
c         Compute the dtau 
c
          if(igrid.eq.1) then
              dtau = ( alpha_a(inu,itcell,ircell) +
     %                 alpha_s(inu,itcell,ircell) ) * ds
          else
              dtau = 0.d0
          endif
c
c         Find albedo and absorb fraction
c
          if(igrid.eq.1) then
              albedo = alpha_s(inu,itcell,ircell)/
     %              (alpha_a(inu,itcell,ircell)+
     %               alpha_s(inu,itcell,ircell))
              absorb = 1.d0-albedo
          endif
c
c         Check if we already reached the end point
c
          if(tau+dtau.gt.taupath) then
c
c             Yes: reached the end point, i.e. reached new event
c
c             Check...
c
              if(igrid.eq.0) stop 94001
c
c             Compute the fraction of the segment we have advanced
c
              fr     = (taupath-tau)/dtau
c
c             Add energy to cell
c
              do ispec=1,dust_nr_species
                  if((dust_quantum(ispec).eq.0).or.
     %                 (iqactive.le.0)) then 
                      do isize=1,dust_nr_size(ispec)
                          cumulener(isize,ispec,itcell,ircell) = 
     %                         cumulener(isize,ispec,itcell,ircell)+
     %                         absorb*(taupath-tau)*energy*
     %                         dust_rho(ispec,itcell,ircell) * 
     %                         kappa_a(inu,isize,ispec) / 
     %                         alpha_a(inu,itcell,ircell)
                      enddo
                  endif
              enddo
c
c             Add photons to scattering source
c
              scatsrc(inu,itcell,ircell) = 
     %             scatsrc(inu,itcell,ircell) +
     %             albedo*(taupath-tau)*energy/
     %             (vol(itcell,ircell)*dust_freq_wgt(inu))
c             
c             Add photons to mean intensity 
c
#             ifdef SAVE_MEANINT
              meanint(inu,itcell,ircell) = 
     %            meanint(inu,itcell,ircell) +
     %            (taupath-tau)*energy/
     %            (vol(itcell,ircell)*dust_freq_wgt(inu)*
     %            (alpha_a(inu,itcell,ircell)+
     %             alpha_s(inu,itcell,ircell))*12.566371d0)
#             endif
c
c             Add photons to mean intensity of 
c             primary (quantum-heating) photons
c
c             NEWSTUFF 12-12-04
c
              if(iqactive.ne.0) then
                  miquant(inu,itcell,ircell) = 
     %                 miquant(inu,itcell,ircell) +
     %                 (taupath-tau)*energy/
     %                 (vol(itcell,ircell)*dust_freq_wgt(inu)*
     %                 (alpha_a(inu,itcell,ircell)+
     %                 alpha_s(inu,itcell,ircell))*12.566371d0)
              endif
c
c             Compute the location of this point
c
c             ... the s
c
              s    = sold + fr * ds
              stot = stot + fr * ds
c
c             ... the r
c
              r   = sqrt(b**2+s**2)
c
c             ... the theta
c
              dummy  = (z0+s*costinf)/r
              theta = acos(dummy)
c
c             ... the mu
c
              mu  = s / r
c
c             ... the phi
c
              dummy  = b**2+s**2-(z0+s*costinf)**2
              if(dummy.le.0.d0) stop 81153
              dummy  = b * sqrt(dummy)
              sinphi = ( b**2 * costinf - z0*s ) / dummy
              if(phi.gt.0.d0) then
                  phi  = asin(sinphi)
              else
                  phi  = pi - asin(sinphi)
              endif
              call period(phi)
c
c             Write path element
c
#             ifdef WRITE_PATH
              write(5,*) r,theta,mu,phi,s,tau_s,stot,energy
#             endif
c
              return
          endif
c
c         Update stot
c
          stot = stot + ds
c
c         Add energy to cell
c
          if(igrid.eq.1) then
c
c             Add energy to cell
c
              do ispec=1,dust_nr_species
                  if((dust_quantum(ispec).eq.0).or.
     %                 (iqactive.le.0)) then 
                      do isize=1,dust_nr_size(ispec)
                          cumulener(isize,ispec,itcell,ircell) = 
     %                         cumulener(isize,ispec,itcell,ircell) +
     %                         absorb*dtau*energy*
     %                         dust_rho(ispec,itcell,ircell) * 
     %                         kappa_a(inu,isize,ispec) / 
     %                         alpha_a(inu,itcell,ircell)
                      enddo
                  endif
              enddo
c
c             Add photons to scattering source
c
              scatsrc(inu,itcell,ircell) = scatsrc(inu,itcell,ircell) +
     %                albedo*dtau*energy/
     %                (vol(itcell,ircell)*dust_freq_wgt(inu))
c
c             Add photons to mean intensity 
c
#             ifdef SAVE_MEANINT
              meanint(inu,itcell,ircell) = 
     %              meanint(inu,itcell,ircell) +
     %              dtau*energy/
     %              (vol(itcell,ircell)*dust_freq_wgt(inu)*
     %              (alpha_a(inu,itcell,ircell)+
     %               alpha_s(inu,itcell,ircell))*12.566371d0)
#             endif
c
c             Add photons to mean intensity of 
c             primary (quantum-heating) photons
c             
c             NEWSTUFF 12-12-04
c             
              if(iqactive.ne.0) then
                  miquant(inu,itcell,ircell) = 
     %                  miquant(inu,itcell,ircell) +
     %                 dtau*energy/
     %                 (vol(itcell,ircell)*dust_freq_wgt(inu)*
     %                 (alpha_a(inu,itcell,ircell)+
     %                  alpha_s(inu,itcell,ircell))*12.566371d0)
              endif
          endif
c
c         Write path element
c
#         ifdef WRITE_PATH
          write(5,*) r,theta,mu,phi,s,tau_s,stot,energy
#         endif
c
c         If we are _at_ the equator, and imirror, then need to flip phi
c
          if((imirt.eq.1).and.(theta.eq.pihalf)) then
              if(phi.lt.pi) stop 64327
              if(idir.ne.2) stop 64328
              if(ileftright.ne.1) stop 64329
              phi = twopi - phi
              ileftright = -1
          endif
c
c         Now let's see to which cell we are going next
c
c         Safety thing:
c
          ircell = -10001
          itcell = -10001
c
c         Go through all four scenarios:
c
          if(idir.eq.1) then
c
c             We are currently on an R-interface
c
              if(ileftright.eq.1) then
                  if(ir1.eq.nr+1) then
c
c                     We are leaving the grid permanently
c
                      costinf = mu * cos(theta) + sqrt(1.d0-mu*mu) * 
     %                          sin(theta) * sin(phi) 
                      r       = 1d30
                      theta   = acos(costinf)
                      if(theta.gt.pihalf) then 
                          theta = pi - theta
                      endif
                      mu      = 1.d0
                      phi     = 0.d0
                      return
                  endif
c
c                 Do an internal consistency check
c                 [CLEANING ERROR MESSAGES 05.08]
c
                  if((ir1.le.0).or.(ir1.gt.nr)) stop 83761
c                  if((it1.le.0).or.(it1.gt.nt)) stop 83762
                  if((it1.le.0).or.(it1.gt.nt)) then
                      if(it1.eq.nt+1) then
                          theta=pihalf-1d-12
                          it1=nt
                      else
                          theta=ttigrid(1)+1d-12
                          it1=1
                      endif
                  endif
c
c                 We are moving from an R-interface to a cell
c                 at larger radii
c
                  r0     = rrigrid(ir1)
                  r1     = rrigrid(ir1+1)
                  t0     = ttigrid(it1)
                  t1     = ttigrid(it1+1)
                  ircell = ir1
                  itcell = it1
              else
c
c                 Do an internal consistency check
c                 [CLEANING ERROR MESSAGES 05.08]
c
                  if((ir1.le.0).or.(ir1.gt.nr+1)) stop 83771
c                  if((it1.le.0).or.(it1.gt.nt)) stop 83772
                  if((it1.le.0).or.(it1.gt.nt)) then
                      if(it1.eq.nt+1) then
                          theta=pihalf-1d-12
                          it1=nt
                      else
                          theta=ttigrid(1)+1d-12
                          it1=1
                      endif
                  endif

c
c                 We are moving from R-interface downwards...
c
                  if(ir1.gt.1) then
c
c                     ...to the previous cell
c
                      r0 = rrigrid(ir1-1)
                  else
c
c                     ...into the inner vacuum
c
                      r0 = 1d-10*rrigrid(1)
                  endif
                  r1     = rrigrid(ir1)
                  t0     = ttigrid(it1)
                  t1     = ttigrid(it1+1)
                  ircell = ir1-1
                  itcell = it1
              endif
          else
c
c             We are currently on a Theta-interface
c
              if(ileftright.eq.1) then
c
c                 Do an internal consistency check
c
                  if((ir1.le.-1).or.(ir1.gt.nr)) stop 83781
                  if((it1.le.0).or.(it1.gt.nt)) stop 83782
c
c                 Moving towards the equator
c
                  if(it1.eq.nteq) stop 7129
                  if(ir1.lt.0) stop 71291
                  r1     = rrigrid(ir1+1)
                  if(ir1.eq.0) then
                      r0 = 1d-10 * r1
                  else
                      r0 = rrigrid(ir1)
                  endif
                  t0     = ttigrid(it1)
                  t1     = ttigrid(it1+1)
                  ircell = ir1
                  itcell = it1
              else
c
c                 Do an internal consistency check
c
                  if((ir1.le.-1).or.(ir1.gt.nr)) stop 83791
                  if((it1.le.0).or.(it1.gt.nt+1)) stop 83792
c
c                 Moving towards the pole
c
                  if(ir1.lt.0) stop 71292
                  r1     = rrigrid(ir1+1)
                  if(ir1.eq.0) then
                      r0 = 1d-10 * r1
                  else
                      r0 = rrigrid(ir1)
                  endif
                  if(it1.gt.1) then
                      t0 = ttigrid(it1-1)
                  else
                      t0 = 1.d-10
                  endif
                  t1     = ttigrid(it1)
                  ircell = ir1
                  itcell = it1-1
              endif
          endif
c
c         Check...
c
          if((ircell.le.-10000).or.(itcell.le.-10000)) stop 33377
          if((r.lt.r0).or.(r.gt.r1).or.
     %       (theta.lt.t0).or.(theta.gt.t1)) then
              write(*,*) 'ERROR in Monte Carlo: cell wrong...'
              write(*,*) idir,ileftright,it1
              write(*,*) r,r0,r1
              write(*,*) theta,t0,t1
              write(*,*) ttigrid(it1)
              stop 13
          endif
c
c         Find out if we are on the grid
c
          igrid = 1
          if(ircell.lt.1) then
              igrid=0
          endif
          if(ircell.gt.nr) then
              igrid=0
              stop 87275
          endif
          if(itcell.lt.1) then
              igrid=0
          endif
          if(itcell.gt.nt) then
              igrid=0
              stop 87276
          endif
c
c         Check cell...
c
          if(igrid.ne.0) then
              r_cell = 0.5d0 * ( rrigrid(ircell) + rrigrid(ircell+1) )
              t_cell = 0.5d0 * ( ttigrid(itcell) + ttigrid(itcell+1) )
              if((r_cell.lt.r0).or.(r_cell.gt.r1).or.
     %           (t_cell.lt.t0).or.(t_cell.gt.t1)) then
                  write(*,*) 'ERROR: Cell wrong...'
                  stop 33732
              endif
          endif
c
c         Increase the cell count
c         (NEWSTUFF 27.01.07)
c
          if(igrid.ne.0) then
              if(iphotcurr.gt.ilastphot(itcell,ircell)) then
                  iphotcount(itcell,ircell) = 
     %               iphotcount(itcell,ircell) + 1
                  ilastphot(itcell,ircell)  = iphotcurr
              endif
          endif
c
c         Increase the tau_s
c
          tau = tau + dtau
c
c         Move new position to old position
c
          rold   = r
          told   = theta
          muold  = mu
          phiold = phi
          sold   = s
c
c         Find the next crossing with a cell interface
c
          call find_next_crossing(rold,r0,r1,told,t0,t1,
     %         muold,phiold,sdum,r,theta,mu,phi,s,
     %         idir,ileftright,b,costinf,z0)
c
c         Check...
c
          if(r.lt.0.d0) then
c              write(*,*) '(error at position B)'
c              write(0,*) '(error at position B)'
              return
          endif
cc##########################
c          if(r.gt.rrigrid(nr+1)) then
c              write(*,*) 'A2 = ',is,r,rrigrid(nr+1)
c          endif
cc##########################
          if(r.gt.rrigrid(nr+1)) then
              write(*,*) 'R out of bounds'
              write(*,*) r,rrigrid(1),rrigrid(nr+1)
              stop 49001
          endif
          if((theta.gt.ttigrid(nt+1)).or.(theta.lt.ttigrid(1))) then
              write(*,*) 'Theta out of bounds'
              write(*,*) theta,ttigrid(1),ttigrid(nt+1)
              write(*,*) rold,told,muold,phiold
              write(*,*) r,theta,mu,phi
              write(*,*) r0,r1,t0,t1,t1/pihalf-1.d0
              stop 49002
          endif
          if((idir.eq.1).and.
     %         ((theta.eq.ttigrid(nt+1)).or.(theta.eq.ttigrid(1)))) then
              write(*,*) 'Funny thing happening with next crossing (1)'
              write(*,*) r,rrigrid(1),rrigrid(2),
     %                   rrigrid(nr),rrigrid(nr+1)
              write(*,*) theta,ttigrid(1),ttigrid(2),
     %                   ttigrid(nt),ttigrid(nt+1)
              stop 49003
          endif
          if((idir.eq.2).and.
     %         ((r.eq.rrigrid(nr+1)).or.(r.eq.rrigrid(1)))) then
              write(*,*) 'Funny thing happening with next crossing (2)'
              write(*,*) r,rrigrid(1),rrigrid(2),
     %                   rrigrid(nr),rrigrid(nr+1)
              write(*,*) theta,ttigrid(1),ttigrid(2),
     %                   ttigrid(nt),ttigrid(nt+1)
              stop 49004
          endif
          if((abs(sold/sdum)-1.d0).gt.1.d-3) stop 59232
          if(s.lt.sold) then
              write(*,*) s,sold,abs((s-sold)/s)
              stop 59233
          endif
c     
      enddo
c
c     If arrived here, then error.
c      
      write(*,*) 'ERROR in Monte Carlo: Should not have arrived here...'
      stop 4444
c
      end


c     --------------------------------------------------------------
c                FIND THE NEXT CELL BOUNDARY CROSSING
c     --------------------------------------------------------------
      subroutine find_next_crossing(r,r0,r1,theta,t0,t1,
     %                  mu,phi,s,r_next,th_next,mu_next,
     %                  phi_next,s_next,idir,ileftright,
     %                  b,costinf,z0)
      implicit none
      doubleprecision r,r0,r1,theta,t0,t1
      doubleprecision mu,phi,s,r_next,th_next,s_next
      doubleprecision mu_next,phi_next
      doubleprecision r_th,th_th,s_th,mu_th,phi_th
      doubleprecision r_r,th_r,s_r,mu_r,phi_r
      doubleprecision dummy,dum,bdum,cdum,sinphi,costh,qdum
      doubleprecision b,costinf,z0,sol1,sol2,ssol1,ssol2
      integer idir,ileftright,ilr_r,ilr_th
c
      doubleprecision pi,pihalf,twopi,fourpi
      parameter(pi=3.14159265358979323846264338328d0)
      parameter(pihalf=1.57079632679489661923132169164d0)
      parameter(twopi=6.28318530717958647692528676656d0)
      parameter(fourpi=12.5663706143591729538505735331d0)
c
c     First check if we are at upper quadrant
c
      if(theta.gt.pihalf) then
          stop 92341
      endif
      if(t1.gt.pihalf) then
          stop 92342
      endif
c
c     Check order
c
      if(r1.lt.r0) then
          stop 92343
      endif
      if(t1.lt.t0) then
          stop 92344
      endif
      if((r.lt.r0).or.(r.gt.r1)) then
          stop 92345
      endif
      if((theta.lt.t0).or.(theta.gt.t1)) then
          write(*,*) t0,t1,theta
          stop 92346
      endif
c
c     Check whether moving inside cell, and not leaving cell...
c
      if((r.eq.r0).and.(mu.lt.0.d0)) stop 77211
      if((r.eq.r1).and.(mu.gt.0.d0)) stop 77212
      if((theta.eq.t0).and.(phi.lt.pi)) then
          write(*,*) '=====',r,r0,r1,theta,t0,t1,mu,phi
          stop 77213
      endif
      if((theta.eq.t1).and.(phi.gt.pi)) stop 77214
c
c     Reset
c
      sol1 = 0.d0
      sol2 = 0.d0
c
c     Find the parameters of the ray, and the value of s
c
      s        = r*mu
      b        = r * sqrt(1.d0-mu*mu)
      costinf  = mu * cos(theta) + sqrt(1.d0-mu*mu) * 
     %                 sin(theta) * sin(phi) 
      z0       = r * ( (1.d0-mu*mu) * cos(theta) - 
     %                 mu * sqrt(1.d0-mu*mu) * sin(theta) * sin(phi) ) 
c
c     Find the next R crossing
c
      if(mu.lt.0.d0) then
c
c         Moving towards smaller radius
c
          if(r0.ge.b) then
c
c             Indeed reaches this smaller radial shell
c
              r_r   = r0
              mu_r  = -sqrt(1.d0-(b/r_r)**2)
              ilr_r = -1
          else
c
c             Moving towards smaller radii, but reaching extremum before
c             reaching the next radial shell, so moving back up.
c
              r_r   = r1
              mu_r  = sqrt(1.d0-(b/r_r)**2)
              ilr_r = 1
          endif
      else
c
c         Moving towards larger radius
c
          r_r   = r1
          mu_r  = sqrt(1.d0-(b/r_r)**2)
          ilr_r = 1
      endif
c
c     Compute the other coordinates of this point...
c
c     ...the s
c
      s_r    = r_r*mu_r
c
c     ...the phi
c
      dummy  = b**2+s_r**2-(z0+s_r*costinf)**2
      if(dummy.le.0.d0) stop 81053
      dummy  = b * sqrt(dummy)
      sinphi = ( b**2 * costinf - z0*s_r ) / dummy
      if(phi.gt.0.d0) then
          phi_r  = asin(sinphi)
      else
          phi_r  = pi - asin(sinphi)
      endif
      call period(phi_r)
c
c     ...and the Theta
c
      th_r  = acos( ( z0 + s_r * costinf ) / r_r )
c
c     Now find the next Theta crossing
c
      if(phi.lt.pi) then
c
c         Moving towards the pole
c
          costh  = cos(t0)
          dum    = costh**2 - costinf**2
          if(dum.eq.0.d0) stop 83218
          bdum   = - 2 * z0 * costinf / dum
          cdum   = ( (b*costh)**2 - z0**2 ) / dum
          qdum   = bdum**2 - 4.d0*cdum
          if(qdum.ge.0.d0) then
c
c             Yes, there are two crossings
c
              sol1   = 0.5d0 * ( -bdum - sqrt(qdum) )
              sol2   = 0.5d0 * ( -bdum + sqrt(qdum) )
c
c             Check if there are solutions in the first place
c
              if(sol2.gt.s) then
c
c                 Get the closest sol after s
c
                  if(sol1.lt.s) then
                      s_th = sol2
                  else
                      s_th = sol1
                  endif
                  th_th  = t0
                  ilr_th = -1
              else
c
c                 All solutions are on other quadrant
c                 So: no theta crossing ahead.
c
                  s_th = 1d30
c
c                 Check something...
c
                  if((abs(costinf).gt.cos(t0)).or.
     %               (abs(costinf).lt.cos(t1))) then
                      write(*,*) 'Monte Carlo: Internal consistency ',
     %                     'failure'
                      stop 44333
                  endif
c
c                 Since we know it is an R crossing, go to 200
c
                  goto 200
              endif
          else
c
c             No, the extremum is before the next Theta gridline,
c             so we move back towards the equator. 
c
c             First check if it1 is the equator
c
              if(t1.eq.pihalf) then
c
c                 Next point is equator. Needs special treatment.
c
                  s_th   = - z0/costinf
                  if((abs((s_th-s)/s).gt.1.d-8).and.(s_th.gt.s)) then
c
c                     Yes, indeed, the next point is -z0/costinf
c
                      th_th  = t1
                      ilr_th = 1
                  else
c
c                     s_th is current point. So: no solution...
c
                      s_th = 1d30
                      goto 200
                  endif
              else
c
c                 Here we have a potential numerical problem: 
c                 since we move back to a theta-line where we came from 
c                 (unless we came from an R-line), one of the solutions will 
c                 be the current point. To avoid this, we compute the crossing 
c                 with theta+1D-6.
c
                  costh  = cos(t1+COSTEPS)
                  dum    = costh**2 - costinf**2
                  if(dum.eq.0.d0) stop 83218
                  bdum   = - 2 * z0 * costinf / dum
                  cdum   = ( (b*costh)**2 - z0**2 ) / dum
                  qdum   = bdum**2 - 4.d0*cdum
                  if(qdum.lt.0.d0) stop 61321
                  ssol1  = 0.5d0 * ( -bdum - sqrt(qdum) )
                  ssol2  = 0.5d0 * ( -bdum + sqrt(qdum) )
c
                  costh  = cos(t1)
                  dum    = costh**2 - costinf**2
                  if(dum.eq.0.d0) stop 83218
                  bdum   = - 2 * z0 * costinf / dum
                  cdum   = ( (b*costh)**2 - z0**2 ) / dum
                  qdum   = bdum**2 - 4.d0*cdum
                  if(qdum.lt.0.d0) stop 61322
                  sol1   = 0.5d0 * ( -bdum - sqrt(qdum) )
                  sol2   = 0.5d0 * ( -bdum + sqrt(qdum) )
c
c                 Find the solution
c
                  if(ssol1.gt.s) then
c
c                     sol1 is the solution
c
                      s_th   = sol1
                      th_th  = t1
                      ilr_th = 1
                  elseif(ssol2.gt.s) then
c
c                     sol2 is the solution
c
                      s_th   = sol2
                      th_th  = t1
                      ilr_th = 1
                  else
c
c                     There is no solution: So: no theta crossing ahead.
c
                      s_th = 1d30
c
c                     Check something...
c
                      if((abs(costinf).gt.cos(t0)).or.
     %                     (abs(costinf).lt.cos(t1+COSTEPS))) then
                          write(*,*) 'Monte Carlo: Internal ',
     %                         'consistency failure'
                          write(*,*) cos(t0),cos(t1+COSTEPS),costinf
                          stop 44332
                      endif
c
c                     Since we know it is an R crossing, go to 200
c
                      goto 200
                  endif
              endif
          endif
      else
c
c         Moving towards the equator
c
          if(t1.eq.pihalf) then
c
c             Next point is equator. Needs special treatment.
c
              s_th   = - z0/costinf
ccccc              if((abs((s_th-s)/s).gt.1.d-8).and.(s_th.gt.s)) then
ccccc              >>>> BUGFIX 31-10-02 <<<<
ccccc                NOTE: If one day this code will be adapted to 
ccccc                no-mirror-symmetry in Theta, then this line must
ccccc                be reconsidered, since it is now assumed that if
ccccc                we arrive here, then phi points upwards.
              if(s_th.gt.s) then
c
c                 Yes, indeed, the next point is -z0/costinf
c
                  th_th  = t1
                  ilr_th = 1
              else
c
c                 s_th is current point. So: no solution...
c
                  s_th = 1d30
                  goto 200
              endif
          else
c
c             Regular points get normal treatment
c
              costh  = cos(t1)
              dum    = costh**2 - costinf**2
              if(dum.eq.0.d0) stop 83218
              bdum   = - 2 * z0 * costinf / dum
              cdum   = ( (b*costh)**2 - z0**2 ) / dum
              qdum   = bdum**2 - 4.d0*cdum
              if(qdum.lt.0.d0) then
                  write(*,*) r,theta,mu,phi
                  write(*,*) z0,costh,costinf
                  write(*,*) bdum**2,4.d0*cdum,qdum
                  stop 99538
              endif
              sol1   = 0.5d0 * ( -bdum - sqrt(qdum) )
              sol2   = 0.5d0 * ( -bdum + sqrt(qdum) )
              if(sol2.gt.s) then
c
c                 As expected, there is a solution. Now find out which.
c
                  if(sol1.lt.s) then
                      s_th = sol2
                  else
                      s_th = sol1
                  endif
                  th_th  = t1
                  ilr_th = 1
              else
c
c                 There is no solution: So: no theta crossing ahead.
c                 This can happen if we were just at an R-crossing,
c                 while in the theta-extremum loop.
c
                  s_th = 1d30
c
c                 Check something...
c
                  if((abs(costinf).gt.cos(t0)).or.
     %                 (abs(costinf).lt.cos(t1))) then
                      write(*,*) 'Monte Carlo: Internal consistency ',
     %                     'failure'
                      stop 44331
                  endif
c
c                 Since we know it is an R crossing, go to 200
c
                  goto 200
              endif
          endif
      endif
c
c     Compute the coordinates of this s...
c
c     ...the r
c
      r_th   = sqrt( b**2 + s_th**2 )
c
c     ...the mu
c
      mu_th  = s_th / r_th
c
c     ...the phi
c
      dummy  = b**2+s_th**2-(z0+s_th*costinf)**2
      if(dummy.le.0.d0) stop 81053
      dummy  = b * sqrt(dummy)
      sinphi = ( b**2 * costinf - z0*s_th ) / dummy
      if(phi.gt.0.d0) then
          phi_th  = asin(sinphi)
      else
          phi_th  = pi - asin(sinphi)
      endif
      call period(phi_th)
c
c     Check which crossing comes first
c
 200  continue
      if(s_r.lt.s_th) then
          r_next     = r_r
          th_next    = th_r 
          mu_next    = mu_r 
          phi_next   = phi_r
          s_next     = s_r  
          idir       = 1
          ileftright = ilr_r
      else
          r_next     = r_th
          th_next    = th_th 
          mu_next    = mu_th
          phi_next   = phi_th
          s_next     = s_th
          idir       = 2
          ileftright = ilr_th
      endif
c
c     Check
c     [CLEANING ERROR MESSAGES 05.08]
c      
      if(abs((s_next-s)/s).lt.MINIEPS) then
c          write(*,*) 'Bizarre chance incident... (773218). ',
c     %               'Skipping 1 photon.'
c          write(*,*) abs((s_next-s)/s)
c          write(*,*) s,s_next,s_r,sol1,sol2
c         write(*,*) 'Skipping 1 photon due to photon-in-corner event.'
          r_next  = -1d30
          return
      endif
      if(s_next.lt.s) then
c          write(*,*) 'Internal error (773219). Skipping 1 photon.'
c          write(0,*) 'Internal error (773219). Skipping 1 photon.'
c          write(*,*) 'S    : ',s_next,s_r,s_th,idir
c          write(*,*) ssol1,ssol2,s
c          write(*,*) abs((s_th-s)/s)
c          write(*,*) 'R    : ',r_next,r0,r1
c          write(*,*) 'Theta: ',th_next,t0,t1
          r_next  = -1d30
          return
      endif
      if((r_next.lt.r0).or.(r_next.gt.r1)) then
c          write(*,*) 'Internal error (773210). Skipping 1 photon.'
c          write(0,*) 'Internal error (773210). Skipping 1 photon.'
c          write(*,*) r_next,r0,r1
          r_next  = -1d30
          return
      endif
      if((th_next.lt.t0).or.(th_next.gt.t1)) then
c          write(*,*) 'Internal error (773211). Skipping 1 photon.'
c          write(0,*) 'Internal error (773211). Skipping 1 photon.'
c          write(*,*) th_next,t0,t1
          r_next  = -1d30
          return
      endif
c
      end


c     --------------------------------------------------------------
c                         DO ABSORPTION EVENT
c
c     This subroutine increases the temperature of a cell according
c     to the mass it contains. And it randomly creates a new inucur.
c     --------------------------------------------------------------
      subroutine do_absorption_event(vol,rhodust,kappa_a,alpha_a,tdust,
     %                               cumulener,cumulener_bk,enthres,
     %                               freqdistr,energy,ir,it,inucur,
     %                               imethod,ifast,iseed,iqactive,
     %                               idestroy)
      implicit none
      integer ir,it,inucur,iseed,imethod,ifast,iqactive,idestroy
      doubleprecision vol(FRSIZE_Y_SMALL,FRSIZE_X),energy,enthres
      doubleprecision alpha_a(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision kappa_a(FRSIZE_FREQ,DUST_SIZE_MAX,
     %                        DUST_SPECIES_MAX)
      doubleprecision rhodust(DUST_SPECIES_MAX,
     %                      FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision tdust(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                      FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision cumulener(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision cumulener_bk(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision freqdistr(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
c
#include "common_grid.h"
#include "common_boundary.h"
#include "common_dust.h"
#include "common_montecarlo.h"
c
      doubleprecision pi,pihalf,twopi,fourpi
      parameter(pi=3.14159265358979323846264338328d0)
      parameter(pihalf=1.57079632679489661923132169164d0)
      parameter(twopi=6.28318530717958647692528676656d0)
      parameter(fourpi=12.5663706143591729538505735331d0)
c
      integer inu,isize,ispec
      doubleprecision bplanck1,bplanck2
      doubleprecision zzbrent,absevfunc,dummy,dum,dum1,dum2
      doubleprecision emisold,tnew,tmin,tmax,tprec,rhodusttot
      doubleprecision cumen,tempav,ran2,rn,fracquant
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX)
      doubleprecision enerpart(DUST_SIZE_MAX,DUST_SPECIES_MAX)
      external absevfunc
      logical error
c
      common/abshandle/fnu_old,fnu_diff,cellalpha,ptclener,cellvol
      doubleprecision fnu_old(FRSIZE_FREQ),fnu_diff(FRSIZE_FREQ)
      doubleprecision cellalpha(FRSIZE_FREQ)
      doubleprecision cellvol,ptclener
c
      logical calctemp
      doubleprecision compute_dusttemp_energy_bd
      doubleprecision compute_dusttemp_coupled_bd
c
      if(dust_freq_wgt(1).eq.0.d0) stop 32222
      if((imethod.eq.0).or.(imethod.eq.1)) then
          write(*,*) 'ERROR: imethod=0 and imethod=1 are obsolete'
          stop
      endif
c
c     Default
c
      idestroy = 0
c
c     First determine how the energy is going to be divided over
c     the dust species and sizes
c
      if(inucur.ge.1) then 
c
c         The normal way: dependent on frequency of incoming photon
c
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  enerpart(isize,ispec) = energy * 
     %                                rhodust(ispec,it,ir) * 
     %                                kappa_a(inucur,isize,ispec) / 
     %                                alpha_a(inucur,it,ir)
              enddo
          enddo
c
c         If quantum heating active:
c         Original method:
c            If the photon is absorbed by a quantum heated grain, then 
c            destroy photon (because the quantum heating is done elsewhere)
c         New method (NEW QUANTUM: 28.04.06):
c            The photon is re-emitted according to the quantum source
c
          if(iqactive.eq.1) then
c
c             ORIGINAL METHOD
c
c             Determine what is the chance that the photon is absorbed
c             by a quantum-heated grain
c
              fracquant = 0.d0
              do ispec=1,dust_nr_species
                  if(dust_quantum(ispec).ne.0) then
                      do isize=1,dust_nr_size(ispec)
                          fracquant = fracquant + enerpart(isize,ispec)
                      enddo
                  endif
              enddo
              fracquant = fracquant / energy
c
c             SMALL BUGFIX (06.07.05): 1.d0 --> 1.d0+1d-10
c
              if(fracquant.gt.1.d0+MINIEPS) stop 92091
c
c             Now pick a random number and determine what happens.
c             If absorbed by a quantum heated grain, then destroy the
c             photon.
c             
              rn = ran2(iseed)
              if(rn.le.fracquant) then
                  idestroy = 1
                  return
              endif
          endif
      elseif(inucur.eq.-1) then
c
c         This is primary energy, and inucur.eq.-1 means that this
c         energy is divided over the species by mass... This only
c         works if dust_nr_size is always 1
c
c         NEWSTUFF 06.06.04
c
          rhodusttot=0.d0
          do ispec=1,dust_nr_species
              rhodusttot = rhodusttot + rhodust(ispec,it,ir)
          enddo
          do ispec=1,dust_nr_species
              if(dust_nr_size(ispec).gt.1) then
                  write(*,*) 'ERROR: This qplus mode only possible'
                  write(*,*) '       with dust_nr_size.eq.1'
                  stop
              endif
              enerpart(1,ispec) = energy * 
     %                     rhodust(ispec,it,ir) / rhodusttot
          enddo
      else
          write(*,*) 'INTERNAL ERROR: Dont know what to do with ',
     %               'inucur = ',inucur
          stop
      endif
c
c     Stupidity check
c
      dum = 0.d0
      do ispec=1,dust_nr_species
          do isize=1,dust_nr_size(ispec)
              dum = dum + enerpart(isize,ispec)
          enddo
      enddo
      if(abs(dum/energy-1.d0).gt.1.d-6) stop 34123
c
c     Determine if the dust temperature(s) and the frequency distribution
c     for this cell must be recomputed
c
      if(ifast.eq.0) then
c
c         For the original method, always recompute temperature and
c         freq dist
c
          calctemp = .true.
      else
c
c         For the new, fast method, check if the energy step is too
c         high.
c
          calctemp = .false.
          if(itempdecoup.eq.1)  then
c
c             If temperatures are decoupled...
c
              do ispec=1,dust_nr_species
                  if((dust_quantum(ispec).eq.0).or.
     %                 (iqactive.le.0)) then 
                      do isize=1,dust_nr_size(ispec)
                          if(abs((cumulener(isize,ispec,it,ir)+
     %                         enerpart(isize,ispec))/
     %                         cumulener_bk(isize,ispec,it,ir)-1.d0)
     %                         .gt.enthres) then
                              calctemp = .true.
                          endif
                      enddo
                  endif
              enddo
          else
c
c             If temperatures are coupled...
c     
              dum1 = 0.d0
              dum2 = 0.d0
              dum  = 0.d0
              do ispec=1,dust_nr_species
                  if((dust_quantum(ispec).eq.0).or.
     %                 (iqactive.le.0)) then 
                      do isize=1,dust_nr_size(ispec)
                          dum1 = dum1 + cumulener(isize,ispec,it,ir)
                          dum2 = dum2 + enerpart(isize,ispec)
                          dum  = dum  + cumulener_bk(isize,ispec,it,ir)
                      enddo
                  endif
              enddo
              if(abs((dum1+dum2)/dum-1.d0).gt.enthres) then
                  calctemp = .true.
              endif
          endif
      endif
c
c     Determine the temperature from the local energy
c     (this need not be hugely accurate)
c
      if(calctemp) then
          if(itempdecoup.eq.1) then
c
c             If temperatures are decoupled...
c
              do ispec=1,dust_nr_species
                  if((dust_quantum(ispec).eq.0).or.
     %                 (iqactive.le.0)) then 
                      do isize=1,dust_nr_size(ispec)
                          cumen = cumulener(isize,ispec,it,ir) /
     %                                 ( dust_rho(ispec,it,ir) *
     %                                   vol(it,ir) )
                          temp(isize,ispec) = 
     %                         compute_dusttemp_energy_bd(
     %                         cumen,ispec,isize)
                          tdust(isize,ispec,it,ir) = temp(isize,ispec)
                      enddo
                  endif
              enddo
          else
c
c             If temperatures are coupled...
c
c             NOTE: Quantum-heated grains are included in the
c                   temperature determination below, even though
c                   the present photon has not been absorbed by one
c                   of them. This is because all grain temperatures
c                   must remain coupled.
c
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      tempav = compute_dusttemp_coupled_bd(
     %                     cumulener(1,1,it,ir),
     %                     dust_rho(1,it,ir),vol(it,ir))
                      temp(isize,ispec)        = tempav
                      tdust(isize,ispec,it,ir) = tempav
                  enddo
              enddo
          endif
      else
c
c         The temperature can still be taken to be the old one
c
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  temp(isize,ispec) = tdust(isize,ispec,it,ir)
              enddo
          enddo
      endif
c
c     Now pick a new frequency, according to dB_nu(T)/dT
c
c     NOTE: This must be enerpart and not cumulener, because
c           suppose we have two dust species, with spec 2 having 0 opacity
c           in the optical. If an optical photon is absorbed, it must
c           have been by spec 1, and thus also spec 1 must re-emit!
c
      call pick_randomfreq_db(tdust(1,1,it,ir),
     %             enerpart,inucur,iseed)
c
c     If the temperature(s) and the freqdist have been recomputed,
c     then we must reset the backup cumulative energy to the present
c     one
c
      if(calctemp) then
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  cumulener_bk(isize,ispec,it,ir) =
     %                 cumulener(isize,ispec,it,ir)
              enddo
          enddo
      endif
c
c     Now reset the iqactive, since from now on all the events for
c     this photon package are thermal (i.e. no quantum-heating
c     events anymore). ONLY IN ORIGINAL PAH METHOD. NEW QUANTUM: 28.04.06
c
      if(iqactive.eq.1) then 
          iqactive = 0
      endif
c
c     Done...
c
      end


c     --------------------------------------------------------------
c               HELPER FUNCTION FOR THE ABSORPTION EVENT
c
c     This function returns 0 for that temperature where the new
c     spectrum of the cell equals the old one plus the energy of
c     the photonj package.
c     --------------------------------------------------------------
      function absevfunc(temp)
      implicit none
      doubleprecision absevfunc,temp
c
#include "common_grid.h"
#include "common_boundary.h"
#include "common_dust.h"
c
      doubleprecision pi,pihalf,twopi,fourpi
      parameter(pi=3.14159265358979323846264338328d0)
      parameter(pihalf=1.57079632679489661923132169164d0)
      parameter(twopi=6.28318530717958647692528676656d0)
      parameter(fourpi=12.5663706143591729538505735331d0)
c
      common/abshandle/fnu_old,fnu_diff,cellalpha,ptclener,cellvol
      doubleprecision fnu_old(FRSIZE_FREQ),fnu_diff(FRSIZE_FREQ)
      doubleprecision cellalpha(FRSIZE_FREQ)
      doubleprecision cellvol,ptclener,bplanck3
c
      integer inu
      doubleprecision demis,fnu
c
      demis = 0.d0
      do inu=1,freq_nr
          fnu           = cellalpha(inu) * bplanck3(temp,freq_nu(inu))
          fnu_diff(inu) = fnu - fnu_old(inu)
          demis         = demis + fnu_diff(inu) * dust_freq_wgt(inu)
      enddo
      demis = demis * fourpi * cellvol
      absevfunc = demis - ptclener
      return
      end



c     --------------------------------------------------------------
c             CHOOSE RANDOM FREQUENCY FROM GIVEN SPECTRUM
c
c     Given a spectrum F_nu this routine will regard F_nu as a
c     probability distribution and choose a frequency randomly
c     within this distribution.
c     --------------------------------------------------------------
      subroutine randomfreq(flux,inupick,iseed)
      implicit none
      doubleprecision flux(FRSIZE_FREQ)
      integer inupick,iseed
c
#include "common_grid.h"
#include "common_dust.h"
c      
      doubleprecision cumul(FRSIZE_FREQ+1),ran2,rn
      integer inu
c
c     Do check
c
      if(dust_freq_wgt(1).eq.0.d0) then
          stop 8331
      endif
c
c     Make cumulative and normalize
c
      cumul(1) = 0.d0
      do inu=1,freq_nr
          cumul(inu+1) = cumul(inu) + flux(inu) * dust_freq_wgt(inu)
      enddo
      do inu=1,freq_nr+1
          cumul(inu) = cumul(inu) / cumul(freq_nr+1)
      enddo
c
c     Get a random number
c
      rn = ran2(iseed)
c
c     Find this in the cumul table
c
      call hunt(cumul,freq_nr,rn,inupick)
c
      end


c     ======================================================================
c                  FIND POSITION IN THE CELL
c
c     When the grid position is choosen the position in the cell has to be 
c     determined.
c     ======================================================================

      subroutine find_position_in_cell(ttigrid,rrigrid,it,ir,
     &                                  r,theta,iseed)
      implicit none

      integer it,ir,nr,nt,iseed
      doubleprecision r,theta
      doubleprecision ttigrid(FRSIZE_Y_SMALL+1)
      doubleprecision rrigrid(FRSIZE_X+1)
      doubleprecision ran2,rn
c
c     Find the theta position
c
c     First pick a random number
c
      rn=ran2(iseed)
c
c     Then find the position
c
      theta=ttigrid(it)+rn*(ttigrid(it+1) - ttigrid(it))
c
c     Find the r position     
c
c     Pick a random number again
c
      rn=ran2(iseed)
c
c     Then find the position
c
      r=rrigrid(ir)+rn*(rrigrid(ir+1) - rrigrid(ir))
c
      end



c     --------------------------------------------------------------
c              HENYEY-GREENSTEIN RANDOM SCATTERING
c
c     This routine gives the random mu_scat and phi_scat according
c     to the small-angle-scattering formula of Henyey-Greenstein.
c
c                1         1 - g^2
c        P(mu) = - -------------------------
c                2 (1 + g^2 - 2 g mu )^(3/2)
c
c     We do this by introducing a variable xi
c
c                1 1 - g  /        1 + g                  \
c        xi   := - -----  | ------------------------- - 1 |
c                2   g    \ (1 + g^2 - 2 g mu )^(1/2)     /
c
c     NOTE: The current implementation has been tested in a simple
c           IDL program called hengreen.pro
c     --------------------------------------------------------------
      subroutine henyeygreenstein(g,mu,phi,iseed)
      implicit none
      doubleprecision g,mu,phi,xi,g2,ran2
      doubleprecision pi
      parameter(pi=3.14159265358979323846264338328d0)
      integer iseed
c
      if(g.ne.0.d0) then
          g2 = g*g
          xi = ran2(iseed)
          mu = (0.5d0/g)*(1.+g2-((1.-g2)/(1.-g+2*g*xi))**2)
      else
          mu = 2.d0*ran2(iseed)-1.d0
      endif
      phi = 2*pi*ran2(iseed)-0.5*pi
c
      end


c     --------------------------------------------------------------
c             ROTATE SCATTERING DIRECTION INTO REAL DIRECTION
c
c     This is a stupid-man's version of a 3-D rotation. Instead of
c     using standard formulae, I use matrix rotation. 
c     --------------------------------------------------------------
      subroutine rotate(mu_in,phi_in,mu_scat,phi_scat,mu_out,phi_out)
      implicit none
      doubleprecision mu_in,phi_in,mu_scat,phi_scat,mu_out,phi_out
c
      doubleprecision vecscat(3),vecdum(3),veclab(3),dum
      doubleprecision pi,pihalf,twopi,fourpi
      parameter(pi=3.14159265358979323846264338328d0)
      parameter(pihalf=1.57079632679489661923132169164d0)
      parameter(twopi=6.28318530717958647692528676656d0)
      parameter(fourpi=12.5663706143591729538505735331d0)
c
c     First create the scattering direction in the local particle
c     frame
c
      dum        = sqrt(1.d0-mu_scat**2)
      vecscat(1) = mu_scat
      vecscat(2) = dum * cos(phi_scat)
      vecscat(3) = dum * sin(phi_scat)
c
c     Now rotate vector over mu (=3-axis)
c
      dum        = sqrt(1.d0-mu_in**2)
      vecdum(1)  = mu_in*vecscat(1) - dum*vecscat(2)
      vecdum(2)  = dum*vecscat(1) + mu_in*vecscat(2)
      vecdum(3)  = vecscat(3)
c
c     Now rotate vector around Z (=1) axis, so that the current
c     direction vector lies in the (1-2) plane
c
      veclab(1)  = vecdum(1)
      veclab(2)  = cos(phi_in)*vecdum(2) - sin(phi_in)*vecdum(3)
      veclab(3)  = sin(phi_in)*vecdum(2) + cos(phi_in)*vecdum(3)
c
c     Renormalize
c
      dum        = sqrt(veclab(1)**2+veclab(2)**2+veclab(3)**2)
      veclab(1)  = veclab(1)/dum
      veclab(2)  = veclab(2)/dum
      veclab(3)  = veclab(3)/dum
c
c     Now convert back to mu,phi
c
      mu_out     = veclab(1)
      if(veclab(2).ne.0.d0) then
          phi_out    = atan(veclab(3)/veclab(2))
          if(veclab(2).lt.0.d0) then
              phi_out = phi_out+pi
          endif
          if(phi_out.lt.0.d0) then
              phi_out = phi_out + 2.d0*pi
          endif
      else
          if(veclab(3).gt.0.d0) then 
              phi_out = pi/2.d0
          else
              phi_out = 3.d0*pi/2.d0
          endif
      endif
c
      end


c     --------------------------------------------------------------
c                THE BLACKBODY PLANCK FUNCTION B_nu(T)
c
c     This function computes the Blackbody function 
c
c                    2 h nu^3 / c^2
c        B_nu(T)  = -------------------    [ erg / cm^2 s ster Hz ]
c                   exp(-h nu / kT) - 1
c
c     ARGUMENTS:
c        nu    [Hz]            = Frequency
c        temp  [K]             = Electron temperature
c     --------------------------------------------------------------
      function bplanck1(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanck1
c
      if(temp.eq.0.d0) then 
          bplanck1 = 0.d0
          return
      endif
c
      bplanck1 = 1.47455d-47 * nu * nu * nu / 
     %      (exp(4.7989d-11 * nu / temp)-1.d0) + 1.d-290
c
      return
      end

      function bplanck2(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanck2
c
      if(temp.eq.0.d0) then 
          bplanck2 = 0.d0
          return
      endif
c
      bplanck2 = 1.47455d-47 * nu * nu * nu / 
     %      (exp(4.7989d-11 * nu / temp)-1.d0) + 1.d-290
c
      return
      end

      function bplanck3(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanck3
c
      if(temp.eq.0.d0) then 
          bplanck3 = 0.d0
          return
      endif
c
      bplanck3 = 1.47455d-47 * nu * nu * nu / 
     %      (exp(4.7989d-11 * nu / temp)-1.d0) + 1.d-290
c
      return
      end



c     --------------------------------------------------------------
c                     FIX ANGLE WITHIN [0,2PI>
c     --------------------------------------------------------------
      subroutine period(phi)
      implicit none
      doubleprecision phi,pi
      parameter(pi=3.14159265358979323846264338328d0)
 103  continue
      if(phi.lt.0.d0) then
          phi = phi + 2.d0*pi
          goto 103
      endif
 104  continue
      if(phi.ge.2.d0*pi) then
          phi = phi - 2.d0*pi
          goto 104
      endif
      end


c     --------------------------------------------------------------
c                          MAKE SAFETY BACKUP
c     --------------------------------------------------------------
      subroutine make_safety_backup(ievenodd,nphot,nr,nt,iphot,
     %                imethod,iseed,scatsrc,miquant,
     %                cumulener,cumulener_bk,
     %                freqdistr,tdust,integerspec,count,ifast)
      implicit none
c
      logical ievenodd
      integer nphot,nr,nt,iphot,imethod,iseed,count,ifast
      doubleprecision scatsrc(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision miquant(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision cumulener(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision cumulener_bk(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision freqdistr(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision tdust(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                      FRSIZE_Y_SMALL,FRSIZE_X)
      integer integerspec(FRSIZE_FREQ,FRSIZE_Y_SMALL)
c
      integer inu,ir,it,isize,ispec
      integer iformat
      character*80 filename
c
      doubleprecision aa,ran2
      integer ii
      integer NTAB
      parameter(NTAB=32)
      common/randbk/bkinit,bkidum,bkidum2,bkiy,bkiv
      integer bkinit,bkidum,bkidum2,bkiy,bkiv(NTAB)
c
#include "common_grid.h"
#include "common_montecarlo.h"
#include "common_dust.h"
c
c     Set format number
c
      iformat=1
      if(ifast.ne.0) then
          iformat=2
      endif
c
c     Retrieve random number settings
c     (NOTE: Using added option of ran2; see nrecip.F)
c
      bkinit=-123456
      aa=ran2(iseed)
c
c     Determine to which file to write this safety save
c
      if(ievenodd) then
          filename='radmc_save_e.dat'
      else
          filename='radmc_save_o.dat'
      endif
c
c     Open and write the safety backup
c     
      open(unit=1,file=filename,status='unknown',form='unformatted')
      write(FRMT) iformat              ! Format number
      write(FRMT) nphot,nr,nt,freq_nr  ! Array sizes
      write(FRMT) iphot,count          ! Where were we? (photon number)
      write(FRMT) bkidum,bkidum2,bkiy, ! Current settings of random generator
     %            (bkiv(ii),ii=1,NTAB)     
      write(FRMT) imethod              ! Method index
      do ir=1,nr                    ! The scattering source array
          do it=1,nt
              do inu=1,freq_nr
                  write(FRMT) scatsrc(inu,it,ir)
              enddo
          enddo
      enddo
      if(iquantum.ne.0) then
          do ir=1,nr                ! The quantum-heating radiation field
              do it=1,nt
                  do inu=1,freq_nr
                      write(FRMT) miquant(inu,it,ir)
                  enddo
              enddo
          enddo
      endif
      do ir=1,nr                    ! The cumulative energy array
          do it=1,nt
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      write(FRMT) cumulener(isize,ispec,it,ir)      
                  enddo
              enddo
          enddo
      enddo
      if(iformat.eq.2) then
          do ir=1,nr            ! The backup cumulative energy array
              do it=1,nt
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          write(FRMT) cumulener(isize,ispec,it,ir)      
                      enddo
                  enddo
              enddo
          enddo
          do ir=1,nr            ! The backup cumulative energy array
              do it=1,nt
                  do inu=1,freq_nr
                      write(FRMT) freqdistr(inu,it,ir)
                  enddo
              enddo
          enddo
      endif
      do ir=1,nr                    ! The tdust array
          do it=1,nt
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      write(FRMT) tdust(isize,ispec,it,ir)      
                  enddo
              enddo
          enddo
      enddo
      do it=1,nt                    ! The ineteger spectrum
          do inu=1,freq_nr
              write(FRMT) integerspec(inu,it)
          enddo
      enddo
      close(1)
c
c     Write to which file the safety backup was made
c
      open(unit=1,file='radmc_save.info')
      if(ievenodd) then
          write(1,*) 'e'
      else
          write(1,*) 'o'
      endif
      write(1,*) iphot,nphot
      close(1)
c
c     Flip ievenodd
c
      ievenodd = .not.ievenodd
c
      end


c     --------------------------------------------------------------
c                          IMPORT SAFETY BACKUP
c     --------------------------------------------------------------
      subroutine read_safety_backup(ievenodd,nphot,nr,nt,iphot,
     %                imethod,iseed,scatsrc,miquant,
     %                cumulener,cumulener_bk,
     %                freqdistr,tdust,integerspec,count)
      implicit none
c
      logical ievenodd
      integer nphot,nr,nt,iphot,imethod,iseed,count
      doubleprecision scatsrc(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision miquant(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision cumulener(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision cumulener_bk(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                          FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision freqdistr(FRSIZE_FREQ,FRSIZE_Y_SMALL,FRSIZE_X)
      doubleprecision tdust(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                      FRSIZE_Y_SMALL,FRSIZE_X)
      integer integerspec(FRSIZE_FREQ,FRSIZE_Y_SMALL)
c
      integer inu,ir,it,isize,ispec
      character*80 filename
      character char
      logical fex
      integer iformat
c
      doubleprecision aa,ran2
      integer ii
      integer NTAB
      parameter(NTAB=32)
      common/randbk/bkinit,bkidum,bkidum2,bkiy,bkiv
      integer bkinit,bkidum,bkidum2,bkiy,bkiv(NTAB)
c
#include "common_grid.h"
#include "common_montecarlo.h"
#include "common_dust.h"
c
c     Write to which file the safety backup was made
c
      inquire(file='radmc_save.info',exist=fex)
      if(.not.fex) then 
          write(*,*) 'Could not find file radmc_save.info'
          write(*,*) 'Cannot restart simulation'
          stop 13
      endif
      open(unit=1,file='radmc_save.info',status='old')
      read(1,*) char
      if(char.eq.'e') then
          ievenodd=.true.
          filename='radmc_save_e.dat'
      elseif(char.eq.'o') then
          ievenodd=.false.
          filename='radmc_save_o.dat'
      else
          write(*,*) '****************************************'
          write(*,*) '----------------------------------------'
          write(*,*) '----     RUN IS ALREADY FINISHED    ----'
          write(*,*) '----       RESTART UNNECESSARY      ----'
          write(*,*) '----------------------------------------'
          write(*,*) ' (If you wish to restart anyway, then '
          write(*,*) '  you must put an e or o at the first '
          write(*,*) '  line of the radmc_save.info file)'
          write(*,*) ' ' 
          write(*,*) ' ' 
          close(1)
          stop ' '
      endif
      close(1)
c      write(*,*) 'Reading safety dump ',filename
c
c     Open and write the safety backup
c     
      inquire(file=filename,exist=fex)
      if(.not.fex) then 
          write(*,*) 'Could not find file ',filename
          write(*,*) 'Cannot restart simulation'
          stop 13
      endif
      open(unit=1,file=filename,status='old',form='unformatted')
      read(FRMT) iformat              ! Format number
      read(FRMT) nphot,nr,nt,freq_nr  ! Array sizes
      read(FRMT) iphot,count          ! Where were we? (photon number)
      read(FRMT) bkidum,bkidum2,bkiy, ! Current settings of random generator
     %            (bkiv(ii),ii=1,NTAB)     
      read(FRMT) imethod              ! Method index
      do ir=1,nr                    ! The scattering source array
          do it=1,nt
              do inu=1,freq_nr
                  read(FRMT) scatsrc(inu,it,ir)
              enddo
          enddo
      enddo
      if(iquantum.ne.0) then
          do ir=1,nr                ! The quantum-heating mean intensity
              do it=1,nt
                  do inu=1,freq_nr
                      read(FRMT) miquant(inu,it,ir)
                  enddo
              enddo
          enddo
      endif
      do ir=1,nr                    ! The cumulative energy array
          do it=1,nt
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      read(FRMT) cumulener(isize,ispec,it,ir)
                  enddo
              enddo
          enddo
      enddo
      if(iformat.eq.2) then
          do ir=1,nr            ! The backup cumulative energy array
              do it=1,nt
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          read(FRMT) cumulener_bk(isize,ispec,it,ir)
                      enddo
                  enddo
              enddo
          enddo
          do ir=1,nr            ! The backup cumulative energy array
              do it=1,nt
                  do inu=1,freq_nr
                      read(FRMT) freqdistr(inu,it,ir)
                  enddo
              enddo
          enddo
      endif
      do ir=1,nr                    ! The tdust array
          do it=1,nt
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      read(FRMT) tdust(isize,ispec,it,ir)      
                  enddo
              enddo
          enddo
      enddo
      do it=1,nt                    ! The ineteger spectrum
          do inu=1,freq_nr
              read(FRMT) integerspec(inu,it)
          enddo
      enddo
      close(1)
c
c     Fix random number settings
c     (NOTE: Using added option of ran2; see nrecip.F)
c
      bkinit=123456
      aa=ran2(iseed)
c
c     Flip ievenodd
c
      ievenodd = .not.ievenodd
c
      end




c     ==============================================================
c                   NEW ROUTINES FOR THE OPTIMIZATION
c                              JANUARY 2007
c     ==============================================================


c     --------------------------------------------------------------
c                       MAKE TEMPERATURE GRID
c
c     For the databases we need a temperature grid. Make it here.
c     --------------------------------------------------------------
      subroutine make_temperature_grid(ntemp,temp0,temp1)
      implicit none
c
      integer ntemp
      doubleprecision temp0,temp1
c
#include "common_montecarlo.h"
c
      integer itemp
c
      if(ntemp.gt.DB_NTEMP_MAX) then
          write(*,*) 'ERROR: The temperatere grid could not be'
          write(*,*) '   made: too many ntemp.'
      endif
c
      db_ntemp = ntemp
      do itemp=1,ntemp
          db_temp(itemp) = temp 0 * (temp1/temp0)**((itemp-1.d0)/
     %                              (ntemp-1.d0))
      enddo
c     
      end



c     --------------------------------------------------------------
c             COMPUTE THE ENERGY BELONGING TO A TEMPERATURE
c
c     This routine computes the energy output per gram of dust
c     of a given species and size for an array of temperatures.
c     --------------------------------------------------------------
      subroutine make_emiss_dbase()
      implicit none
c
#include "common_grid.h"
#include "common_montecarlo.h"
#include "common_montecarlo_arrays.h"
#include "common_dust.h"
c
      common/abshandle/fnu_old,fnu_diff,cellalpha,ptclener,cellvol
      doubleprecision fnu_old(FRSIZE_FREQ),fnu_diff(FRSIZE_FREQ)
      doubleprecision cellalpha(FRSIZE_FREQ)
      doubleprecision cellvol,ptclener
c
      integer itemp,inu,ispec,isize
      doubleprecision absevfunc,diffemis(FRSIZE_FREQ)
      doubleprecision cumul(FRSIZE_FREQ+1)
c     
c     A loop over all species and sizes
c
      do ispec=1,dust_nr_species
          do isize=1,dust_nr_size(ispec)
c
c             Compute the emissivity and the total emitted energy
c             per second per gram dust of this species/size.
c
              ptclener = 0.d0
              cellvol  = 1.d0
              do inu=1,freq_nr
                  cellalpha(inu) = kappa_a(inu,isize,ispec)
                  fnu_old(inu)   = 0.d0
              enddo
              do itemp=1,db_ntemp
                  db_enertemp(itemp,isize,ispec) = 
     %                  absevfunc(db_temp(itemp))
                  do inu=1,freq_nr
                      db_emiss(inu,itemp,isize,ispec) =
     %                     fnu_diff(inu)
                  enddo
              enddo
c
c             Now compute the derivative of the emissivity
c             with respect to energy, so that we can obtain
c             a right distribution function for picking a
c             new photon frequency. We immediately make
c             a cumulative version of this for each
c             species, so that the picking of a new frequency
c             for a photon goes in two steps:
c              1) Pick a species/size from the energies
c              2) Pick, within this selected species, a freq
c
c             NOTE!!!! VERY IMPORTANT!!!: 
c               This is done slightly differently from the
c               original method in which the exact difference
c               of fnu and fnu_old was taken. But this is not
c               too impportant, because 1) the normalization
c               if irrelevant and 2) for sufficient photon
c               statistics the derivative is close enough to
c               the difference that it is OK.
c
c             For the first temperature we simply take the
c             emissivity function itself
c
              itemp = 1
              do inu=1,freq_nr
                  diffemis(inu) = db_emiss(inu,itemp,isize,ispec)
              enddo
              cumul(1) = 0.d0
              do inu=1,freq_nr
                  cumul(inu+1) = cumul(inu) + 
     %                diffemis(inu) * dust_freq_wgt(inu)
              enddo
              do inu=1,freq_nr+1
                  db_cumulnorm(inu,itemp,isize,ispec) = 
     %                 cumul(inu) / cumul(freq_nr+1)
              enddo
c
c             For the rest take indeed the difference
c              
              do itemp=2,db_ntemp
                  do inu=1,freq_nr
                      diffemis(inu) = 
     %                     db_emiss(inu,itemp,isize,ispec) - 
     %                     db_emiss(inu,itemp-1,isize,ispec) 
                  enddo
                  cumul(1) = 0.d0
                  do inu=1,freq_nr
                      cumul(inu+1) = cumul(inu) + 
     %                     diffemis(inu) * dust_freq_wgt(inu)
                  enddo
                  do inu=1,freq_nr+1
                      db_cumulnorm(inu,itemp,isize,ispec) = 
     %                     cumul(inu) / cumul(freq_nr+1)
                  enddo
              enddo
          enddo
      enddo
c
      end



c     --------------------------------------------------------------
c               COMPUTE THE TEMPERATURE FOR A GIVEN ENERGY
c
c     This routine computes the dust temperature for a given 
c     energy-per-gram-of-dust for this species.
c     --------------------------------------------------------------
      function compute_dusttemp_energy_bd(ener,ispec,isize)
      implicit none
c
      integer ispec,isize
      doubleprecision ener,compute_dusttemp_energy_bd
c
#include "common_grid.h"
#include "common_montecarlo.h"
#include "common_dust.h"
c
      integer itemp
      doubleprecision eps,temp
c
c     First find the energy in the grid
c
      call hunt(db_enertemp(1,isize,ispec),db_ntemp,ener,itemp)
c
c     Check if we are in range
c
      if(itemp.ge.db_ntemp) then
          write(*,*) 'ERROR: Too high temperature discovered' 
          stop 76823
      endif
      if(itemp.le.0) then
c
c         Temperature presumably below lowest temp in dbase
c
          eps = ener/db_enertemp(1,isize,ispec)
c#######################
          if(eps.gt.1.d0) stop 9911
c#######################
          temp = eps*db_temp(1)
      else
c
c         Temperature in the right range: default action
c
          eps = (ener-db_enertemp(itemp,isize,ispec)) / 
     %          (db_enertemp(itemp+1,isize,ispec) - 
     %           db_enertemp(itemp,isize,ispec))
c#######################
          if((eps.gt.1.d0).or.(eps.lt.0.d0)) stop 9912
c#######################
          temp = (1.d0-eps)*db_temp(itemp) +
     %           eps*db_temp(itemp+1)
      endif
c
c     Return
c
      compute_dusttemp_energy_bd = temp
      return
c
      end


c     --------------------------------------------------------------
c               COMPUTE THE TEMPERATURE FOR A GIVEN ENERGY
c
c     This routine computes the dust temperature for a given 
c     energy-per-gram-of-dust for this species.
c     --------------------------------------------------------------
      function compute_dusttemp_coupled_bd(energy,rho,vol)
      implicit none
c
      integer ispec,isize
      doubleprecision vol
      doubleprecision energy(DUST_SIZE_MAX,DUST_SPECIES_MAX)
      doubleprecision rho(DUST_SPECIES_MAX)
      doubleprecision compute_dusttemp_coupled_bd
c
#include "common_grid.h"
#include "common_montecarlo.h"
#include "common_dust.h"
c
      integer itemp
      doubleprecision eps,temp,entot,dummy1,dummy2
      doubleprecision en1,en2
c
c     Compute the total energy per gram dust
c
      entot = 0.d0
      do ispec=1,dust_nr_species
          do isize=1,dust_nr_size(ispec)
              entot  = entot  + energy(isize,ispec)
          enddo
      enddo      
      entot = entot / vol
c
c     First find the energy in the grid
c
      call hunt_temp(rho,entot,itemp)
c
c     Check if we are in range
c
      if(itemp.ge.db_ntemp) then
          write(*,*) 'ERROR: Too high temperature discovered' 
          stop 76823
      endif
      if(itemp.le.0) then
c
c         Temperature presumably below lowest temp in dbase
c
          en2 = 0.d0
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  en2 = en2 + rho(ispec)*
     %                     db_enertemp(1,isize,ispec)
              enddo
          enddo
          eps = entot/en2
c#######################
          if(eps.gt.1.d0) stop 59911
c#######################
          temp = eps*db_temp(1)
      else
c
c         Temperature in the right range: default action
c
          en1 = 0.d0
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  en1 = en1 + rho(ispec)*
     %                     db_enertemp(itemp,isize,ispec)
              enddo
          enddo
          en2 = 0.d0
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  en2 = en2 + rho(ispec)*
     %                     db_enertemp(itemp+1,isize,ispec)
              enddo
          enddo
          eps = (entot-en1) / (en2-en1)
c#######################
          if((eps.gt.1.d0).or.(eps.lt.0.d0)) stop 99124
c#######################
          temp = (1.d0-eps)*db_temp(itemp) +
     %           eps*db_temp(itemp+1)
      endif
c
c     Return
c
      compute_dusttemp_coupled_bd = temp
      return
c
      end


c     --------------------------------------------------------------
c               COMPUTE NEW IFREQ FROM LOCAL SITUATION
c     --------------------------------------------------------------
      subroutine pick_randomfreq_db(temp,enerpart,inupick,iseed)
      implicit none
c     
      integer inupick,iseed
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX)
      doubleprecision enerpart(DUST_SIZE_MAX,DUST_SPECIES_MAX)
c
#include "common_grid.h"
#include "common_montecarlo.h"
#include "common_dust.h"
c
      doubleprecision enercum(DUST_SIZE_MAX+1,DUST_SPECIES_MAX)
      doubleprecision enercumm(DUST_SPECIES_MAX+1)
      doubleprecision cumul(FRSIZE_FREQ+1)
      doubleprecision ran2,rn,eps
      integer inu,ispec,isize,itemp
c
c     Figure out which species and size to take. This can 
c     be determined from the energies
c
      enercumm(1) = 0.d0
      do ispec=1,dust_nr_species
          enercum(1,ispec) = 0.d0
          do isize=1,dust_nr_size(ispec)
              enercum(isize+1,ispec) = enercum(isize,ispec) + 
     %                               enerpart(isize,ispec)
          enddo
c         BUGFIX 06.08.07
c          enercumm(ispec+1) = enercumm(ispec) + 
c     %              enercum(dust_nr_size(ispec),ispec)
          enercumm(ispec+1) = enercumm(ispec) + 
     %              enercum(dust_nr_size(ispec)+1,ispec)
      enddo
c     BUGFIX 09.08.07
      if(dust_nr_species.gt.1) then 
          rn = ran2(iseed)*enercumm(dust_nr_species+1)
          call hunt(enercumm,dust_nr_species,rn,ispec)
c###################
          if((ispec.lt.1).or.(ispec.gt.dust_nr_species)) stop 50209
c###################
      else
          ispec = 1
      endif
      if(dust_nr_size(ispec).gt.1) then 
          rn = ran2(iseed)*enercum(dust_nr_size(ispec)+1,ispec)
          call hunt(enercum(1,ispec),dust_nr_size(ispec),
     %              rn,isize)
c###################
          if((isize.lt.1).or.(isize.gt.dust_nr_size(ispec))) stop 50210
c###################
      else
          isize = 1
      endif
c
c     Find the temperature in the database
c
      call hunt(db_temp,db_ntemp,temp(isize,ispec),itemp)
c
c     Check if we are in range
c
      if(itemp.ge.db_ntemp) then
          write(*,*) 'ERROR: Too high temperature discovered' 
          stop 77823
      endif
      if(itemp.le.0) then
c
c         Temperature presumably below lowest temp in dbase
c
c         BUGFIX 06.08.07-B
c
c          eps = temp(isize,ispec)/db_temp(1)
cc#######################
c          if(eps.gt.1.d0) stop 94911
cc#######################
          itemp = 1
          eps   = 0.d0
      else
c
c         Temperature in the right range: default action
c
          eps = (temp(isize,ispec)-db_temp(itemp)) / 
     %          (db_temp(itemp+1)-db_temp(itemp))
c#######################
          if((eps.gt.1.d0).or.(eps.lt.0.d0)) stop 94941
c#######################
      endif
c
c     Now, within this species/size, find the frequency
c
      do inu=1,freq_nr+1
          cumul(inu) = (1.d0-eps)*db_cumulnorm(inu,itemp,isize,ispec) +
     %                 eps*db_cumulnorm(inu,itemp+1,isize,ispec)
      enddo
c######################
      if(abs(cumul(freq_nr+1)-1.d0).gt.1d-11) stop 43465
c######################
      rn = ran2(iseed)
      call hunt(cumul,freq_nr,rn,inupick)
c######################
      if((inupick.lt.1).or.(inupick.gt.freq_nr)) stop 8189
c######################
c
      end


c     -------------------------------------------------------------
c              HUNT IN TOTAL ENERGY FOR TEMPERATURE
c     (MODIFIED NRECIP ROUTINE)
c        energy = total energy divided by volume
c     -------------------------------------------------------------
      subroutine hunt_temp(rho,energy,jlo)
      integer jlo,n
      doubleprecision energy,xx,x
      doubleprecision rho(DUST_SPECIES_MAX)
      integer inc,jhi,jm
      logical ascnd
c
#include "common_grid.h"
#include "common_montecarlo.h"
#include "common_dust.h"
c
      n = db_ntemp
      x = energy
      ascnd=.true.
      if(jlo.le.0.or.jlo.gt.n)then
        jlo=0
        jhi=n+1
        goto 3
      endif
      inc=1
      xx=0.d0
      do ispec=1,dust_nr_species
          do isize=1,dust_nr_size(ispec)
              xx = xx + rho(ispec) * 
     %          db_enertemp(jlo,isize,ispec)
          enddo
      enddo
      if(x.ge.xx.eqv.ascnd)then
1       jhi=jlo+inc
        xx=0.d0
        do ispec=1,dust_nr_species
            do isize=1,dust_nr_size(ispec)
                xx = xx + rho(ispec) * 
     %               db_enertemp(jhi,isize,ispec)
            enddo
        enddo
        if(jhi.gt.n)then
          jhi=n+1
        else if(x.ge.xx.eqv.ascnd)then
          jlo=jhi
          inc=inc+inc
          goto 1
        endif
      else
        jhi=jlo
2       jlo=jhi-inc
        xx=0.d0
        do ispec=1,dust_nr_species
            do isize=1,dust_nr_size(ispec)
                xx = xx + rho(ispec) * 
     %               db_enertemp(jlo,isize,ispec)
            enddo
        enddo
        if(jlo.lt.1)then
          jlo=0
        else if(x.lt.xx.eqv.ascnd)then
          jhi=jlo
          inc=inc+inc
          goto 2
        endif
      endif
3     if(jhi-jlo.eq.1)return
      jm=(jhi+jlo)/2
      xx=0.d0
      do ispec=1,dust_nr_species
          do isize=1,dust_nr_size(ispec)
              xx = xx + rho(ispec) * 
     %             db_enertemp(jm,isize,ispec)
          enddo
      enddo
      if(x.gt.xx.eqv.ascnd)then
        jlo=jm
      else
        jhi=jm
      endif
      goto 3
      end
C  (C) Copr. 1986-92 Numerical Recipes Software =v1.9"217..


#endif /* INCLUDE_DUST */
      
